
__copyright__ = 'Copyright 2013-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os

from ... import states    as rps
from ... import agent     as rpa
from ... import constants as rpc
from ... import utils     as rpu


# ------------------------------------------------------------------------------
# 'enum' for RP's spawner types
EXECUTING_NAME_POPEN   = 'POPEN'
EXECUTING_NAME_FLUX    = 'FLUX'
EXECUTING_NAME_SLEEP   = 'SLEEP'


# ------------------------------------------------------------------------------
#
class AgentExecutingComponent(rpu.Component):
    '''
    Manage the creation of Task processes, and watch them until they are
    completed (one way or the other).  The spawner thus moves the task from
    PendingExecution to Executing, and then to a final state (or PendingStageOut
    of course).
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        rpu.Component.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Spawner
    #
    @classmethod
    def create(cls, cfg, session):

        # Make sure that we are the base-class!
        if cls != AgentExecutingComponent:
            raise TypeError('Factory only available to base class!')

        name = cfg['spawner']

        from .popen    import Popen
        from .flux     import Flux
        from .sleep    import Sleep

        impl = {
            EXECUTING_NAME_POPEN  : Popen,
            EXECUTING_NAME_FLUX   : Flux,
            EXECUTING_NAME_SLEEP  : Sleep,
        }

        if name not in impl:
            raise ValueError('AgentExecutingComponent %s unknown' % name)

        return impl[name](cfg, session)



    # --------------------------------------------------------------------------
    #
    def initialize(self):

      # self._log.debug('exec base initialize')

        # The spawner/executor needs the ResourceManager information which have
        # been collected during agent startup.
        self._rm = rpa.ResourceManager.create(self._cfg.resource_manager,
                                              self._cfg, self._log, self._prof)

        self._pwd      = os.path.realpath(os.getcwd())
        self.sid       = self._cfg['sid']
        self.resource  = self._cfg['resource']
        self.rsbox     = self._cfg['resource_sandbox']
        self.ssbox     = self._cfg['session_sandbox']
        self.psbox     = self._cfg['pilot_sandbox']
        self.gtod      = '$RP_PILOT_SANDBOX/gtod'
        self.prof      = '$RP_PILOT_SANDBOX/prof'

        # if so configured, let the Task know what to use as tmp dir
        self._task_tmp = self._cfg.get('task_tmp',
                                       os.environ.get('TMP', '/tmp'))


        if self.psbox.startswith(self.ssbox):
            self.psbox = '$RP_SESSION_SANDBOX%s'  % self.psbox[len(self.ssbox):]
        if self.ssbox.startswith(self.rsbox):
            self.ssbox = '$RP_RESOURCE_SANDBOX%s' % self.ssbox[len(self.rsbox):]
        if self.ssbox.endswith(self.sid):
            self.ssbox = '%s$RP_SESSION_ID/'      % self.ssbox[:-len(self.sid)]

        self.register_input(rps.AGENT_EXECUTING_PENDING,
                            rpc.AGENT_EXECUTING_QUEUE, self.work)

        self.register_output(rps.AGENT_STAGING_OUTPUT_PENDING,
                             rpc.AGENT_STAGING_OUTPUT_QUEUE)

        self.register_publisher (rpc.AGENT_UNSCHEDULE_PUBSUB)
        self.register_subscriber(rpc.CONTROL_PUBSUB, self.command_cb)


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        raise NotImplementedError('work is not implemented')


    # --------------------------------------------------------------------------
    #
    def command_cb(self, topic, msg):

        raise NotImplementedError('work is not implemented')


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2016, http://radical.rutgers.edu"
__license__   = "MIT"


import sys
import time
import pymongo

import radical.utils as ru

from ..   import utils     as rpu
from ..   import constants as rpc

from ..db import DBSession



def out(msg):
    sys.stdout.write('%s\n' % msg)
    sys.stdout.flush()


# ------------------------------------------------------------------------------
#
DEFAULT_BULK_COLLECTION_TIME =  1.0  # seconds
DEFAULT_BULK_COLLECTION_SIZE =  100  # seconds


# ------------------------------------------------------------------------------
#
class Update(rpu.Worker):
    '''
    An UpdateWorker pushes Task and Pilot state updates to mongodb.  Its instances
    compete for update requests on the update_queue.  Those requests will be
    triplets of collection name, query dict, and update dict.  Update requests
    will be collected into bulks over some time (BULK_COLLECTION_TIME) and
    number (BULK_COLLECTION_SIZE) to reduce number of roundtrips.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        rpu.Worker.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._sid        = self._cfg['sid']
        self._dburl      = self._cfg['dburl']

        # get db handle from a connected, non-primary session
        self._dbs   = DBSession(self._sid, self._dburl, {}, self._log, connect=True)
        self._coll  = self._dbs._c
        self._bulk  = self._coll.initialize_ordered_bulk_op()
        self._last  = time.time()        # time of last bulk push
        self._uids  = list()             # list of collected uids
        self._lock  = ru.Lock()          # protect _bulk

        self._bulk_time = self._cfg.bulk_time
        self._bulk_size = self._cfg.bulk_size

        self.register_subscriber(rpc.STATE_PUBSUB, self._state_cb)
        self.register_timed_cb(self._idle_cb, timer=self._bulk_time)


    # --------------------------------------------------------------------------
    #
    @classmethod
    def create(cls, cfg, session):

        return cls(cfg, session)


    # --------------------------------------------------------------------------
    #
    def _timed_bulk_execute(self, flush=False):

        # is there anything to execute?
        if not self._uids:
            return True

        now = time.time()
        age = now - self._last

        # only push if flush is forced, or when collection time or size
        # have been exceeded
        if not flush \
           and age < self._bulk_time \
           and len(self._uids) < self._bulk_size:
            return False

        try:
            self._bulk.execute()

        except pymongo.errors.OperationFailure as e:
            self._log.exception('bulk exec error: %s' % e.details)
            raise

        except Exception as e:
            self._log.exception('mongodb error: %s', e)
            raise

        self._prof.prof('update_pushed', msg='bulk size: %d' % len(self._uids))

      # for entry in self._uids:
      #
      #     uid   = entry[0]
      #     state = entry[2]
      #
      #     if state:
      #         self._prof.prof('update_pushed', uid=uid, msg=state)
      #     else:
      #         self._prof.prof('update_pushed', uid=uid)

        # empty bulk, refresh state
        self._last = now
        self._bulk = self._coll.initialize_ordered_bulk_op()
        self._uids = list()

        return True


    # --------------------------------------------------------------------------
    #
    def _idle_cb(self):

        with self._lock:
            self._timed_bulk_execute()

        return True


    # --------------------------------------------------------------------------
    #
    def _state_cb(self, topic, msg):
        '''

        # FIXME: this documentation is not final, nor does it reflect reality!

        'msg' is expected to be of the form ['cmd', 'thing'], where 'thing' is
        an entity to update in the DB, and 'cmd' specifies the mode of update.

        'things' are expected to be dicts with a 'type' and 'uid' field.  If
        either one does not exist, an exception is raised.

        Supported types are:

          - task
          - pilot

        supported 'cmds':

          - delete      : delete can be delayed until bulk is collected/flushed
          - update      : update can be delayed until bulk is collected/flushed
          - state       : update can be delayed until bulk is collected/flushed
                          only state and state history are updated
          - delete_flush: delete is sent immediately (possibly in a bulk)
          - update_flush: update is sent immediately (possibly in a bulk)
          - state_flush : update is sent immediately (possibly in a bulk)
                          only state and state history are updated
          - flush       : flush pending bulk

        The 'thing' can contains '$set' and '$push' fields, which will then be
        used as given.  For all other fields, we use the following convention:

          - scalar values: use '$set'
          - dict   values: use '$set'
          - list   values: use '$push'

        That implies that all potential 'list' types should be defined in the
        initial 'thing' insert as such, as (potentially empty) lists.

        For 'cmd' in ['state', 'state_flush'], only the 'uid' and 'state' fields
        of the given 'thing' are used, all other fields are ignored.  If 'state'
        does not exist, an exception is raised.
        '''

        try:
            cmd    = msg['cmd']
            things = msg['arg']

          # cmds = ['delete',       'update',       'state',
          #         'delete_flush', 'update_flush', 'state_flush', 'flush']
            if cmd not in ['update', 'insert']:
                return True

            if cmd == 'insert':
                self._dbs.insert_tasks(ru.as_list(things))
                return True


            # FIXME: we don't have any error recovery -- any failure to update
            #        state in the DB will thus result in an exception here and tear
            #        down the module.
            for thing in ru.as_list(things):

                # got a new request.  Add to bulk (create as needed),
                # and push bulk if time is up.
                uid   = thing['uid']
                ttype = thing['type']
                state = thing['state']

                if 'clone' in uid:
                    # we don't push clone states to DB
                    return True

              # self._prof.prof('update_request', msg=state, uid=uid)

                if not state:
                    # nothing to push
                    return True

                # create an update document
                update_dict          = dict()
                update_dict['$set']  = dict()
                update_dict['$push'] = dict()

                for key,val in thing.items():
                    # never set _id, states (to avoid index clash, doubled ops)
                    if key not in ['_id', 'states', 'cmds']:
                        update_dict['$set'][key] = val

                # we set state, put (more importantly) we push the state onto
                # the 'states' list, so that we can later get state progression
                # in sync with the state model, even if they have been pushed
                # here out-of-order
                update_dict['$push']['states'] = state

                with self._lock:

                    # push the update request onto the bulk
                    self._uids.append([uid, ttype, state])
                    self._bulk.find  ({'uid' : uid,
                                       'type': ttype}) \
                              .update(update_dict)

            with self._lock:
                # attempt a timed update
                self._timed_bulk_execute()

            return True

        except:
            return False


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2018-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os

from .base import RMInfo, ResourceManager


# ------------------------------------------------------------------------------
#
class LSF(ResourceManager):

    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, rm_info: RMInfo) -> RMInfo:

        # LSF hostfile format:
        #
        #     node_1
        #     node_1
        #     ...
        #     node_2
        #     node_2
        #     ...
        #
        # There are in total "-n" entries (number of tasks of the job)
        # and "-R" entries per node (tasks per host).
        #
        hostfile = os.environ.get('LSB_DJOB_HOSTFILE')
        if not hostfile:
            raise RuntimeError('$LSB_DJOB_HOSTFILE not set')

        smt   = rm_info.threads_per_core
        nodes = self._parse_nodefile(hostfile, smt=smt)

        # LSF adds login and batch nodes to the hostfile (with 1 core) which
        # needs filtering out.
        #
        # It is possible that login/batch nodes were not marked at hostfile
        # and were not filtered out, thus we assume that there is only one
        # such node with 1 core (otherwise assertion error will be raised later)
        # *) affected machine(s): Lassen@LLNL
        filtered = list()
        for node in nodes:
            if   'login' in node[0]: continue
            elif 'batch' in node[0]: continue
            elif 1       == node[1]: continue
            filtered.append(node)

        nodes = filtered

        lsf_cores_per_node = self._get_cores_per_node(nodes)
        if rm_info.cores_per_node:
            assert (rm_info.cores_per_node == lsf_cores_per_node)
        else:
            rm_info.cores_per_node = lsf_cores_per_node

        self._log.debug('found %d nodes with %d cores',
                        len(nodes), rm_info.cores_per_node)

        # While LSF node names are unique and could serve as node uids, we
        # need an integer index later on for resource set specifications.
        # (LSF starts node indexes at 1, not 0)
        rm_info.node_list = self._get_node_list(nodes, rm_info)

        # NOTE: blocked cores should be in sync with SMT level,
        #       as well with CPU indexing type ("logical" vs "physical").
        #       Example of CPU indexing on Summit:
        #          https://github.com/olcf-tutorials/ERF-CPU-Indexing
        #
        # The current approach uses "logical" CPU indexing
        # FIXME: set cpu_indexing as a parameter in resource config

        # node uids need to be indexes starting at 1 in order to be usable for
        # jsrun ERF spec files
        for idx, node in enumerate(rm_info.node_list):
            node['node_id'] = str(idx + 1)

        return rm_info


# ------------------------------------------------------------------------------






import io
import os
import sys
import time
import queue

import threading         as mt
import multiprocessing   as mp

import radical.utils     as ru

from .worker import Worker


# ------------------------------------------------------------------------------
#
class DefaultWorker(Worker):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg=None, session=None):

        # generate a MPI rank dependent UID for each worker process
        # FIXME: this should be delegated to ru.generate_id
        # FIXME: rank determination should be moved to RU
        rank = None

        if rank is None: rank = os.environ.get('PMIX_RANK')
        if rank is None: rank = os.environ.get('PMI_RANK')
        if rank is None: rank = os.environ.get('OMPI_COMM_WORLD_RANK')

        if rank is None: rank = 0
        else           : rank = int(rank)

        # only rank 0 registers with the master
        if rank == 0: register = True
        else        : register = False

        super().__init__(cfg=cfg, session=session, register=register)

        # connect to the master queues
        self._res_put = ru.zmq.Putter('result',  self._cfg.info.res_addr_put)
        self._req_get = ru.zmq.Getter('request', self._cfg.info.req_addr_get,
                                                 cb=self.request_cb)

        # keep worker ID and rank
        self._cfg['rank'] = rank
        self._cfg['uid']  = '%s.%03d' % (self._cfg['uid'], rank)

        self._n_cores = self._cfg.worker_descr.cores_per_rank
        self._n_gpus  = self._cfg.worker_descr.gpus_per_rank

        self._res_evt = mp.Event()          # set on free resources
        self._my_term = mt.Event()          # for start/stop/join

        self._mlock   = ru.Lock(self._uid)  # lock `_modes`
        self._modes   = dict()              # call modes (call, exec, eval, ...)

        # We need to make sure to run only up to `gpn` tasks using a gpu
        # within that pool, so need a separate counter for that.
        self._resources = {'cores' : [0] * self._n_cores,
                           'gpus'  : [0] * self._n_gpus}

      # self._log.debug('cores %s', str(self._resources['cores']))
      # self._log.debug('gpus  %s', str(self._resources['gpus']))

        # resources are initially all free
        self._res_evt.set()

      # # create a multiprocessing pool with `cpn` worker processors.  Set
      # # `maxtasksperchild` to `1` so that we get a fresh process for each
      # # task.  That will also allow us to run command lines via `exec`,
      # # effectively replacing the worker process in the pool for a specific
      # # task.
      # #
      # # We use a `fork` context to inherit log and profile handles.
      # #
      # # NOTE: The mp documentation is wrong; mp.Pool does *not* have a context
      # #       parameters.  Instead, the Pool has to be created within
      # #       a context.
      # ctx = mp.get_context('fork')
      # self._pool = ctx.Pool(processes=self._n_cores,
      #                       initializer=None,
      #                       maxtasksperchild=1)
      # NOTE: a multiprocessing pool won't work, as pickle is not able to
      #       serialize our worker object.  So we use our own process pool.
      #       It's not much of a loss since we want to respawn new processes for
      #       each task anyway (to improve isolation).
        self._pool  = dict()  # map task uid to process instance
        self._plock = ru.Lock('p' + self._uid)  # lock _pool

        # We also create a queue for communicating results back, and a thread to
        # watch that queue
        self._result_queue  = mp.Queue()
        self._result_thread = mt.Thread(target=self._result_watcher)
        self._result_thread.daemon = True
        self._result_thread.start()

        # run worker initialization *before* starting to work on requests.
        # the worker provides three builtin methods:
        #     eval:  evaluate a piece of python code with `eval`
        #     exec:  evaluate a piece of python code with `exec`
        #     call:  execute  a method or function call
        #     proc:  execute  a command line (fork/exec)
        #     shell: execute  a shell command
        self.register_mode('eval',  self._eval)
        self.register_mode('exec',  self._exec)
        self.register_mode('call',  self._call)
        self.register_mode('proc',  self._proc)
        self.register_mode('shell', self._shell)

        self.pre_exec()

        # prepare base env dict used for all tasks
        self._task_env = dict()
        for k,v in os.environ.items():
            if k.startswith('RP_'):
                self._task_env[k] = v


    # --------------------------------------------------------------------------
    #
    def start(self):

        pass


    # --------------------------------------------------------------------------
    #
    def stop(self):

        self._my_term.set()


    # --------------------------------------------------------------------------
    #
    def join(self):

        # FIXME
        while True:
            if self._my_term.wait(1):
                break


    # --------------------------------------------------------------------------
    #
    def pre_exec(self):
        '''
        This method can be overloaded by the Worker implementation to run any
        pre_exec commands before spawning worker processes.
        '''

        pass


    # --------------------------------------------------------------------------
    #
    def register_mode(self, name, executor):

        assert name not in self._modes

        self._modes[name] = executor


    # --------------------------------------------------------------------------
    #
    def _eval(self, data):
        '''
        We expect data to have a single entry: 'code', containing the Python
        code to be eval'ed
        '''

        bak_stdout = sys.stdout
        bak_stderr = sys.stderr

        strout = None
        strerr = None

        try:
            # redirect stdio to capture them during execution
            sys.stdout = strout = io.StringIO()
            sys.stderr = strerr = io.StringIO()

            val = eval(data['code'])
            out = strout.getvalue()
            err = strerr.getvalue()
            ret = 0

        except Exception as e:
            self._log.exception('_eval failed: %s' % (data))
            self._log.exception('_eval failed: %s' % (data))
            val = None
            out = strout.getvalue()
            err = strerr.getvalue() + ('\neval failed: %s' % e)
            ret = 1

        finally:
            # restore stdio
            sys.stdout = bak_stdout
            sys.stderr = bak_stderr


        return out, err, ret, val


    # --------------------------------------------------------------------------
    #
    def _exec(self, data):
        '''
        We expect data to have a single entry: 'code', containing the Python
        code to be exec'ed.  Data can have an optional entry `pre_exec` which
        can be used for any import statements and the like which need to run
        before the executed code.
        '''

        bak_stdout = sys.stdout
        bak_stderr = sys.stderr

        strout = None
        strerr = None

        try:
            # redirect stdio to capture them during execution
            sys.stdout = strout = io.StringIO()
            sys.stderr = strerr = io.StringIO()

            pre  = data.get('pre_exec', '')
            code = data['code']

            # create a wrapper function around the given code
            lines = code.split('\n')
            outer = 'def _my_exec():\n'
            for line in lines:
                outer += '    ' + line + '\n'

            # call that wrapper function via exec, and keep the return value
            src = '%s\n\n%s\n\nresult=_my_exec()' % (pre, outer)

            # assign a local variable to capture the code's return value.
            loc = dict()
            exec(src, {}, loc)                # pylint: disable=exec-used # noqa
            val = loc['result']
            out = strout.getvalue()
            err = strerr.getvalue()
            ret = 0

        except Exception as e:
            self._log.exception('_exec failed: %s' % (data))
            val = None
            out = strout.getvalue()
            err = strerr.getvalue() + ('\nexec failed: %s' % e)
            ret = 1

        finally:
            # restore stdio
            sys.stdout = bak_stdout
            sys.stderr = bak_stderr


        return out, err, ret, val


    # --------------------------------------------------------------------------
    #
    def _call(self, data):
        '''
        We expect data to have a three entries: 'method' or 'function',
        containing the name of the member method or the name of a free function
        to call, `args`, an optional list of unnamed parameters, and `kwargs`,
        and optional dictionary of named parameters.
        '''

        if 'method' in data:
            to_call = getattr(self, data['method'], None)

        elif 'function' in data:
            names   = dict(list(globals().items()) + list(locals().items()))
            to_call = names.get(data['function'])

        else:
            raise ValueError('no method or function specified: %s' % data)

        if not to_call:
            raise ValueError('callable not found: %s' % data)


        args   = data.get('args',   [])
        kwargs = data.get('kwargs', {})

        bak_stdout = sys.stdout
        bak_stderr = sys.stderr

        strout = None
        strerr = None

        try:
            # redirect stdio to capture them during execution
            sys.stdout = strout = io.StringIO()
            sys.stderr = strerr = io.StringIO()

            val = to_call(*args, **kwargs)
            out = strout.getvalue()
            err = strerr.getvalue()
            ret = 0

        except Exception as e:
            self._log.exception('_call failed: %s' % (data))
            val = None
            out = strout.getvalue()
            err = strerr.getvalue() + ('\ncall failed: %s' % e)
            ret = 1

        finally:
            # restore stdio
            sys.stdout = bak_stdout
            sys.stderr = bak_stderr


        return out, err, ret, val


    # --------------------------------------------------------------------------
    #
    def _proc(self, data):
        '''
        We expect data to have two entries: 'exe', containing the executabele to
        run, and `args` containing a list of arguments (strings) to pass as
        command line arguments.  We use `sp.Popen` to run the fork/exec, and to
        collect stdout, stderr and return code
        '''

        try:
            import subprocess as sp

            exe  = data['exe']
            args = data.get('args', list())
            env  = data.get('env',  dict())

            args = '%s %s' % (exe, ' '.join(args))

            proc = sp.Popen(args=args,      env=env,
                            stdin=None,     stdout=sp.PIPE, stderr=sp.PIPE,
                            close_fds=True, shell=True)
            out, err = proc.communicate()
            ret      = proc.returncode

        except Exception as e:
            self._log.exception('popen failed: %s' % (data))
            out = None
            err = 'exec failed: %s' % e
            ret = 1

        return out, err, ret, None


    # --------------------------------------------------------------------------
    #
    def _shell(self, data):
        '''
        We expect data to have a single entry: 'cmd', containing the command
        line to be called as string.
        '''

        try:
            out, err, ret = ru.sh_callout(data['cmd'], shell=True)

        except Exception as e:
            self._log.exception('_shell failed: %s' % (data))
            out = None
            err = 'shell failed: %s' % e
            ret = 1

        return out, err, ret, None


    # --------------------------------------------------------------------------
    #
    # FIXME: an MPI call mode should be added.  That could work along these
    #        lines of:
    #
    # --------------------------------------------------------------------------
    #  def _mpi(self, data):
    #
    #      try:
    #          cmd = rp.agent.launch_method.construct_command(data,
    #                  executable=self.exe, args=data['func'])
    #          out = rp.sh_callout(cmd)
    #          err = None
    #          ret = 0
    #
    #      except Exception as e:
    #          self._log.exception('_mpi failed: %s' % (data))
    #          out = None
    #          err = 'mpi failed: %s' % e
    #          ret = 1
    #
    #      return out, err, ret, None
    # --------------------------------------------------------------------------
    #
    # For that to work we would need to be able to create a LM here, but ideally
    # not replicate the work done in the agent executor.


    # --------------------------------------------------------------------------
    #
    def _alloc_task(self, task):
        '''
        allocate task resources
        '''

        with self._mlock:

            cores = task.get('cores', 1)
            gpus  = task.get('gpus' , 0)

            assert cores >= 1
            assert cores <= self._n_cores
            assert gpus  <= self._n_gpus

            if cores > self._resources['cores'].count(0): return False
            if gpus  > self._resources['gpus' ].count(0): return False

            alloc_cores = list()
            alloc_gpus  = list()

            if cores:
                for n in range(self._n_cores):
                    if not self._resources['cores'][n]:
                        self._resources['cores'][n] = 1
                        alloc_cores.append(n)
                        if len(alloc_cores) == cores:
                            break

            if gpus:
                for n in range(self._n_gpus):
                    if not self._resources['gpus'][n]:
                        self._resources['gpus'][n] = 1
                        alloc_gpus.append(n)
                        if len(alloc_gpus) == gpus:
                            break

            task['slots'] = {'cores': alloc_cores,
                             'gpus' : alloc_gpus}
            return True


    # --------------------------------------------------------------------------
    #
    def _dealloc_task(self, task):
        '''
        deallocate task resources
        '''

        with self._mlock:

            resources = task['slots']

            for n in resources['cores']:
                assert self._resources['cores'][n]
                self._resources['cores'][n] = 0

            for n in resources['gpus']:
                assert self._resources['gpus'][n]
                self._resources['gpus'][n] = 0

            # signal available resources
            self._res_evt.set()

            return True


    # --------------------------------------------------------------------------
    #
    def task_pre_exec(self, task):
        '''
        This method is called upon receiving a new request, and can be
        overloaded to perform any preperatory action before the request is acted
        upon
        '''
        pass


    # --------------------------------------------------------------------------
    #
    def task_post_exec(self, task):
        '''
        This method is called upon completing a request, and can be
        overloaded to perform any cleanup action before the request is reported
        as complete.
        '''
        pass


    # --------------------------------------------------------------------------
    #
    def request_cb(self, tasks):
        '''
        grep call type from tasks, check if methods are registered, and
        invoke them.
        '''

        for task in ru.as_list(tasks):

            task['worker'] = self._uid

            self.task_pre_exec(task)

            try:

                # ok, we have work to do.  Check the requirements to see how
                # many cpus and gpus we need to mark as busy
                while not self._alloc_task(task):

                    # no resource - wait for new resources
                    #
                    # NOTE: this will block smaller tasks from being executed
                    #       right now.  alloc_task is not a proper scheduler,
                    #       after all.
                  # while not self._res_evt.wait(timeout=1.0):
                  #     self._log.debug('req_alloc_wait %s', task['uid'])
                  # #   FIXME: `clear` should be locked
                  # self._res_evt.clear()

                    time.sleep(0.01)

                self._prof.prof('req_start', uid=task['uid'], msg=self._uid)

                # we got an allocation for this task, and can run it, so apply
                # to the process pool.  The callback (`self._result_cb`) will
                # pick the task up on completion and free resources.
                #
                # NOTE: we don't use mp.Pool - see __init__ for details

                env = self._task_env
                env['RP_TASK_ID'] = task['uid']

              # ret = self._pool.apply_async(func=self._dispatch, args=[task],
              #                              callback=self._result_cb,
              #                              error_callback=self._error_cb)
                proc = mp.Process(target=self._dispatch, args=(task, env))
              # proc.daemon = True

                with self._plock:

                    # we need to include `proc.start()` in the lock, as
                    # otherwise we may end up getting the `self._result_cb`
                    # before the pid could be registered in `self._pool`.
                    proc.start()
                    self._pool[proc.pid] = proc
                self._log.debug('applied: %s: %s: %s',
                                task['uid'], proc.pid, self._pool.keys())


            except Exception as e:

                self._log.exception('request failed')

                # free resources again for failed task
                self._dealloc_task(task)

                res = {'req': task['uid'],
                       'out': None,
                       'err': 'req_cb error: %s' % e,
                       'ret': 1}

                self._res_put.put(res)


    # --------------------------------------------------------------------------
    #
    def _dispatch(self, task, env):

        # this method is running in a process of the process pool, and will now
        # apply the task to the respective execution mode.
        #
        # NOTE: application of pre_exec directives may got here

        task['pid'] = os.getpid()

        # apply task env settings
        for k,v in env.items():
            os.environ[k] = v

        for k,v in task.get('environment', {}).items():
            os.environ[k] = v

        # ----------------------------------------------------------------------
        def _dispatch_proc(res_lock):
            # FIXME: do we still need this thread?

            import setproctitle
            setproctitle.setproctitle('rp.dispatch.%s' % task['uid'])

            # make CUDA happy
            # FIXME: assume physical device numbering for now
            if task['slots']['gpus']:
                os.environ['CUDA_VISIBLE_DEVICES'] = \
                             ','.join(str(i) for i in task['slots']['gpus'])

            out, err, ret, val = self._modes[mode](task.get('data'))
            res = [task, str(out), str(err), int(ret), val]

            with res_lock:
                self._result_queue.put(res)
        # ----------------------------------------------------------------------


        ret = None
        try:
          # self._log.debug('dispatch: %s: %d', task['uid'], task['pid'])
            mode = task['mode']
            assert mode in self._modes, 'no such call mode %s' % mode

            tout = task.get('timeout')
            self._log.debug('dispatch with tout %s', tout)

          # result = self._modes[mode](task.get('data'))
          # self._log.debug('got result: task %s: %s', task['uid'], result)
          # out, err, ret, val = result
          # # TODO: serialize `val`?
          # res = [task, str(out), str(err), int(ret), val]
          # self._result_queue.put(res)

            res_lock = mp.Lock()
            dispatcher = mp.Process(target=_dispatch_proc, args=(res_lock,))
            dispatcher.daemon = True
            dispatcher.start()
            dispatcher.join(timeout=tout)

            with res_lock:
                if dispatcher.is_alive():
                    dispatcher.terminate()
                    dispatcher.join()
                    out = None
                    err = 'timeout (>%s)' % tout
                    ret = 1
                    res = [task, str(out), str(err), int(ret), None]
                    self._log.debug('put 2 result: task %s', task['uid'])
                    self._result_queue.put(res)
                    self._log.debug('dispatcher killed: %s', task['uid'])

        except Exception as e:

            self._log.exception('dispatch failed')
            out = None
            err = 'dispatch failed: %s' % e
            ret = 1
            res = [task, str(out), str(err), int(ret), None]
            self._log.debug('put 3 result: task %s', task['uid'])
            self._result_queue.put(res)

        finally:
            # if we kill the process too quickly, the result put above
            # will not make it out, thus make sure the queue is empty
            # first.
            ret = 1
            self._result_queue.close()
            self._result_queue.join_thread()
            sys.exit(ret)
          # os.kill(os.getpid(), signal.SIGTERM)



    # --------------------------------------------------------------------------
    #
    def _result_watcher(self):

        try:
            while True:

                try:
                    res = self._result_queue.get(timeout=0.1)
                    self._log.debug('got   result: %s', res)
                    self._result_cb(res)

                except queue.Empty:
                    pass

        except:
            self._log.exception('queue error')
            raise


    # --------------------------------------------------------------------------
    #
    def _result_cb(self, result):

        try:
            task, out, err, ret, val = result
            self._log.debug('result cb: task %s', task['uid'])

            with self._plock:
                pid  = task['pid']
                del self._pool[pid]

            # free resources again for the task
            self._dealloc_task(task)

            res = {'req': task['uid'],
                   'out': out,
                   'err': err,
                   'ret': ret,
                   'val': val}

            self._res_put.put(res)
            self.task_post_exec(task)
            self._prof.prof('req_stop', uid=task['uid'], msg=self._uid)

        except:
            self._log.exception('result cb failed')
            raise



    # --------------------------------------------------------------------------
    #
    def _error_cb(self, error):

        self._log.debug('error: %s', error)
        raise RuntimeError(error)


    # --------------------------------------------------------------------------
    #
    def test(self, idx, seconds):
        # pylint: disable=reimported
        import time
        print('start idx %6d: %.1f' % (idx, time.time()))
        time.sleep(seconds)
        print('stop  idx %6d: %.1f' % (idx, time.time()))


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import copy

import radical.utils as ru

from .continuous import Continuous

from ... import states    as rps
from ... import constants as rpc


# ------------------------------------------------------------------------------
#
# This is a simple extension of the Continuous scheduler which evaluates the
# `colocate` tag of arriving tasks, which is expected to have the form
#
#   colocate : {'ns'   : <string>,
#               'size' : <int>}
#
# where 'ns' (for namespace) is a bag ID, and 'size' is the number of tasks in
# that bag of tasks that need to land on the same host.  The semantics of the
# scheduler is that, for any given namespace, it will schedule either all tasks
# in that ns at the same time on the same node, or will schedule no task of that
# ns at all.
#
# The dominant use case for this scheduler is the execution of coupled
# applications which exchange data via shared local files or shared memory.
#
# FIXME: - failed tasks cannot yet considered, subsequent tasks in the same ns
#          will be scheduled anyway.
#
class ContinuousColo(Continuous):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        Continuous.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        Continuous._configure(self)

        # a 'bag' entry will look like this:
        #
        #   {
        #      'size': 128,    # number of tasks to expect
        #      'uids': [...]}, # ids    of tasks to be scheduled
        #   }

        self._lock      = ru.RLock()   # lock on the bags
        self._tasks     = dict()       # task registry (we use uids otherwise)
        self._unordered = list()       # IDs of tasks which are not colocated
        self._bags      = dict()       # nothing has run, yet

        self._bag_init  = {'size' : 0,
                           'uids' : list()}


    # --------------------------------------------------------------------------
    # overload the main method from the base class
    def schedule_task(self, task):

        self.advance(task, rps.AGENT_SCHEDULING, publish=True, push=False)

        with self._lock:

            # cache ID int to avoid repeated parsing
            uid      = task['uid']
            descr    = task['description']
            colo_tag = descr.get('tags', {}).get('colocate')

            # tasks w/o order info are handled as usual, and we don't keep
            # any infos around
            if not colo_tag:
              # self._log.debug('no tags for %s', uid)
                self._unordered.append(task)
                return

            # this uniit wants to be ordered - keep it in our registry
            assert uid not in self._tasks, 'duplicated task %s' % uid
            self._tasks[uid] = task

            bag   = colo_tag['bag']
            size  = colo_tag['size']

          # self._log.debug('tags %s: %s : %d', uid, bag, size)

            # initiate bag if needed
            if bag not in self._bags:
                self._bags[bag]         = copy.deepcopy(self._bag_init)
                self._bags[bag]['size'] = size

            else:
                assert size == self._bags[bag]['size'], \
                       'inconsistent bag size'

            # add task to order
            self._bags[bag]['uids'].append(uid)

        # try to schedule known tasks
        self._try_schedule()

        return


    # --------------------------------------------------------------------------
    def _try_schedule(self):
        '''
        Schedule all tasks in self._unordered.  Then for all name spaces,
        check if their `current` order has tasks to schedule.  If not and
        we see `size` tasks are `done`, consider the order completed and go
        to the next one.  Break once we find a BoT which is not completely
        schedulable, either because we did not yet get all its tasks, or
        because we run out of resources to place those tasks.
        '''

        self._log.debug('try schedule')
        scheduled = list()  # list of scheduled tasks

        # FIXME: this lock is very aggressive, it should not be held over
        #        the scheduling algorithm's activity.
        # first schedule unordered tasks (
        with self._lock:

            keep = list()
            for task in self._unordered:

                # attempt to schedule this task (use continuous algorithm)
                if Continuous._try_allocation(self, task):

                    # success - keep it and try the next one
                    scheduled.append(task)

                else:
                    # failure - keep task around
                    keep.append(task)

            # keep only unscheduleed tasks
            self._unordered = keep


        # FIXME: this lock is very aggressive, it should not be held over
        #        the scheduling algorithm's activity.
        with self._lock:

            # now check all bags for eligibility, filter scheduled ones
            to_delete = list()
            for bag in self._bags:

                self._log.debug('try bag %s', bag)

                if self._bags[bag]['size'] < len(self._bags[bag]['uids']):
                    raise RuntimeError('inconsistent bag assembly')

                # if bag is complete, try to schedule it
                if self._bags[bag]['size'] == len(self._bags[bag]['uids']):

                    self._log.debug('try bag %s (full)', bag)
                    if self._try_schedule_bag(bag):

                        self._log.debug('try bag %s (placed)', bag)
                        # scheduling works - push tasks out and erase all traces
                        # of the bag (delayed until after iteration)
                        for uid in self._bags[bag]['uids']:

                            scheduled.append(self._tasks[uid])

                        to_delete.append(bag)

            # delete all bags which have been pushed out
            for bag in to_delete:

                del self._bags[bag]


        # advance all scheduled tasks and push them out
        if scheduled:
            self.advance(scheduled, rps.AGENT_EXECUTING_PENDING,
                         publish=True, push=True)

      # self._log.debug('dump')
      # self._log.debug(pprint.pformat(self._bags))


    # --------------------------------------------------------------------------
    #
    def _try_schedule_bag(self, bag):
        '''
        This methods assembles the requiremets of all tasks in a bag into
        a single pseudo-task.  We ask the cont scheduler to schedule that
        pseudo-task for us.  If that works, we disassemble the resulting
        resource slots and assign them to the bag's tasks again, and declare
        success.
        '''

        self._log.debug('try schedule bag %s ', bag)

        tasks  = [self._tasks[uid] for uid in self._bags[bag]['uids']]
        pseudo = copy.deepcopy(tasks[0])

        pseudo['uid'] = 'pseudo.'

        descr = pseudo['description']
        descr['threading_type']   = rpc.POSIX  # force single node
        descr['ranks']            = 1
        descr['cores_per_rank']   = 1
        descr['gpus_per_rank']    = 0
        descr['gpu_type']         = None

        self._log.debug('try schedule uids  %s ', self._bags[bag]['uids'])
      # self._log.debug('try schedule tasks  %s ', pprint.pformat(tasks))

        for task in tasks:
            td = task['description']
            pseudo['uid'] += task['uid']

            descr['cores_per_rank'] += td['ranks'] * td['cores_per_rank']
            descr['gpus_per_rank']  += td['ranks'] * td['gpus_per_rank']

      # self._log.debug('try schedule pseudo %s ', pprint.pformat(pseudo))

        if not Continuous._try_allocation(self, pseudo):

            # cannot scshedule this pseudo task right now, bag has to wait
            return False

        # we got an allocation for the pseudo task, not dissassemble the slots
        # and assign back to the individual tasks in the bag
        slots = copy.deepcopy(pseudo['slots'])
        cpus  = copy.deepcopy(pseudo['slots']['ranks'][0]['core_map'])
        gpus  = copy.deepcopy(pseudo['slots']['ranks'][0]['gpu_map'])

        slots['ranks'][0]['core_map'] = list()
        slots['ranks'][0]['gpu_map']  = list()

        for task in tasks:

            tslots = copy.deepcopy(slots)
            descr  = task['description']

            for _ in range(descr['threads_per_rank']):
                block = list()
                for _ in range(descr['cores_per_rank']):
                    block.append(cpus.pop(0)[0])
                tslots['ranks'][0]['core_map'].append(block)

            for _ in range(descr['gpus_pre_rank']):

                block = list()
                block.append(gpus.pop(0)[0])
                tslots['ranks'][0]['gpu_map'].append(block)

            task['slots'] = tslots

        return True


    # --------------------------------------------------------------------------
    #
    def schedule_cb(self, topic, msg):
        '''
        This cb gets triggered after some tasks got unscheduled, ie. their
        resources have been freed.  We attempt a new round of scheduling at that
        point.
        '''
        self._try_schedule()

        # keep the cb registered
        return True


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2016-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os

import radical.utils as ru

from .base import RMInfo, ResourceManager


# ------------------------------------------------------------------------------
#
class Slurm(ResourceManager):

    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, rm_info: RMInfo) -> RMInfo:

        nodelist = os.environ.get('SLURM_NODELIST')
        if nodelist is None:
            raise RuntimeError('$SLURM_NODELIST not set')

        # Parse SLURM nodefile environment variable
        node_names = ru.get_hostlist(nodelist)
        self._log.info('found SLURM_NODELIST %s. Expanded to: %s',
                       nodelist, node_names)

        if not rm_info.cores_per_node:
            # $SLURM_CPUS_ON_NODE = Number of physical cores per node
            cpn_str = os.environ.get('SLURM_CPUS_ON_NODE')
            if cpn_str is None:
                raise RuntimeError('$SLURM_CPUS_ON_NODE not set')
            rm_info.cores_per_node = int(cpn_str)

        nodes = [(node, rm_info.cores_per_node) for node in node_names]

        rm_info.node_list = self._get_node_list(nodes, rm_info)

        return rm_info


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

# ------------------------------------------------------------------------------
# we *first* import radical.utils, so that the monkeypatching of the logger has
# a chance to kick in before the logging module is pulled by any other 3rd party
# module, and also to monkeypatch `os.fork()` for the `atfork` functionality
import radical.utils as _ru

# ------------------------------------------------------------------------------
# constants and types
from .states     import *
from .constants  import *


# ------------------------------------------------------------------------------
# import API
from .session                   import Session
from .context                   import Context

from .task_manager              import TaskManager
from .task                      import Task
from .pytask                    import PythonTask
from .task_description          import TaskDescription
from .task_description          import TASK_EXECUTABLE, TASK_FUNCTION
from .task_description          import TASK_EXEC, TASK_EVAL, TASK_PROC, TASK_SHELL
from .task_description          import RAPTOR_MASTER, RAPTOR_WORKER

from .pilot_manager             import PilotManager
from .pilot                     import Pilot
from .pilot_description         import PilotDescription


# ------------------------------------------------------------------------------
# make submodules available -- mostly for internal use
from . import utils
from . import worker
from . import tmgr
from . import pmgr
from . import agent

from .agent  import Agent_0
from .agent  import Agent_n

from .raptor import Master, Worker


# ------------------------------------------------------------------------------
#
# get version info
#
import os as _os

version_short, version_detail, version_base, version_branch, \
        sdist_name, sdist_path = _ru.get_version(_os.path.dirname(__file__))

version = version_short


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__copyright__ = "Copyright 2013-2014, http://radical.rutgers.edu"
__license__   = "MIT"

import os
import radical.pilot as rp

verbose  = os.environ.get('RADICAL_PILOT_VERBOSE', 'REPORT')
os.environ['RADICAL_PILOT_VERBOSE'] = verbose

""" DESCRIPTION: Tutorial 3: Coupled Tasks
For every task A1 and B1 a C1 is started.
"""

# READ: The RADICAL-Pilot documentation:
#   https://radicalpilot.readthedocs.io/en/stable/
#
# Try running this example with RADICAL_PILOT_VERBOSE=debug set if
# you want to see what happens behind the scenes!


# ------------------------------------------------------------------------------
#
if __name__ == "__main__":

    RESOURCE_LABEL = None
    PILOT_CORES    = None
    NUMBER_COUPLES = None
    TASK_A_CORES   = None
    TASK_B_CORES   = None
    TASK_C_CORES   = None
    QUEUE          = None

    # Create a new session. No need to try/except this: if session creation
    # fails, there is not much we can do anyways...
    session = rp.Session()

    # all other pilot code is now tried/excepted.  If an exception is caught, we
    # can rely on the session object to exist and be valid, and we can thus tear
    # the whole RP stack down via a 'session.close()' call in the 'finally'
    # clause...
    try:


        # Add a Pilot Manager. Pilot managers manage one or more Pilots.
        print("Initializing Pilot Manager ...")
        pmgr = rp.PilotManager(session=session)


        # ----- CHANGE THIS -- CHANGE THIS -- CHANGE THIS -- CHANGE THIS ------
        #
        # If you want to run this example on your local machine, you don't have
        # to change anything here.
        #
        # Change the resource below if you want to run on a remote resource.
        # You also might have to set the 'project' to your allocation ID if
        # your remote resource does compute time accounting.
        #
        # A list of preconfigured resources can be found at:
        # https://radicalpilot.readthedocs.io/en/stable/machconf.html#preconfigured-resources
        #
        pdesc = rp.PilotDescription ()
        pdesc.resource = RESOURCE_LABEL
        pdesc.runtime  = 30
        pdesc.cores    = PILOT_CORES
        pdesc.cleanup  = True
        pdesc.queue    = QUEUE

        # submit the pilot.
        print("Submitting  Pilot to Pilot Manager ...")
        pilot = pmgr.submit_pilots(pdesc)

        # Combine the Pilot, the Tasks and a scheduler via
        # a TaskManager object.
        print("Initializing Task Manager ...")
        tmgr = rp.TaskManager (session=session)

        # Add the created Pilot to the TaskManager.
        print("Registering  Pilot with Task Manager ...")
        tmgr.add_pilots(pilot)

        # submit A tasks to pilot job
        taskdesc_list_A = []
        for idx in range(NUMBER_COUPLES):

            # -------- BEGIN USER DEFINED Task 1 DESCRIPTION --------- #
            taskdesc = rp.TaskDescription()
            taskdesc.environment = {"TASK_LIST": "A", "TASK_NO": "%02d" % idx}
            taskdesc.executable  = "/bin/echo"
            taskdesc.arguments   = ['"$TASK_LIST Task with id $TASK_NO"']
            taskdesc.cores       = TASK_A_CORES
            # -------- END USER DEFINED Task 1 DESCRIPTION --------- #

            taskdesc_list_A.append(taskdesc)

        # Submit the previously created Task descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning Tasks to the Pilots.
        print("Submit Tasks 'A' to Task Manager ...")
        task_set_A = tmgr.submit_tasks(taskdesc_list_A)

        # submit B tasks to pilot job
        taskdesc_list_B = []
        for idx in range(NUMBER_COUPLES):

            # -------- BEGIN USER DEFINED Task 2 DESCRIPTION --------- #
            taskdesc = rp.TaskDescription()
            taskdesc.environment = {"TASK_LIST": "B", "TASK_NO": "%02d" % idx}
            taskdesc.executable  = "/bin/echo"
            taskdesc.arguments   = ['"$TASK_LIST Task with id $TASK_NO"']
            taskdesc.cores       = TASK_B_CORES
            # -------- END USER DEFINED Task 2 DESCRIPTION --------- #

            taskdesc_list_B.append(taskdesc)

        # Submit the previously created Task descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning Tasks to the Pilots.
        print("Submit Tasks 'B' to Task Manager ...")
        task_set_B = tmgr.submit_tasks(taskdesc_list_B)


        # ---------------------------------------------------------------------
        print("Waiting for 'A' and 'B' tasks to complete...")
        tmgr.wait_tasks()
        print("Executing 'C' tasks now...")
        # ---------------------------------------------------------------------

        # submit 'C' tasks to pilot job. each 'C' task takes the output of
        # an 'A' and a 'B' task and puts them together.
        taskdesc_list_C = []
        for idx in range(NUMBER_COUPLES):

            # -------- BEGIN USER DEFINED Task 3 DESCRIPTION --------- #
            taskdesc = rp.TaskDescription()
            taskdesc.environment = {"TASK_SET": "C", "TASK_NO": "%02d" % idx}
            taskdesc.executable  = "/bin/echo"
            taskdesc.arguments   = ['"$TASK_SET Task with id $TASK_NO"']
            taskdesc.cores       = TASK_C_CORES
            # -------- END USER DEFINED Task 3 DESCRIPTION --------- #

            taskdesc_list_C.append(taskdesc)

        # Submit the previously created Task descriptions to the
        # PilotManager. This will trigger the selected scheduler to start
        # assigning Tasks to the Pilots.
        print("Submit Tasks 'C' to Task Manager ...")
        task_set_C = tmgr.submit_tasks(taskdesc_list_C)

        # ---------------------------------------------------------------------
        print("Waiting for 'C' tasks to complete...")
        tmgr.wait_tasks()
        print("All tasks completed successfully!")


    except Exception as e:
        # Something unexpected happened in the pilot code above
        print("caught Exception: %s" % e)
        raise

    except (KeyboardInterrupt, SystemExit) as e:
        # the callback called sys.exit(), and we can here catch the
        # corresponding KeyboardInterrupt exception for shutdown.  We also catch
        # SystemExit (which gets raised if the main threads exits for some other
        # reason).
        print("need to exit now: %s" % e)

    finally:
        # always clean up the session, no matter if we caught an exception or
        # not.
        print("closing session")
        session.close ()

        # the above is equivalent to
        #
        #   session.close (cleanup=True, terminate=True)
        #
        # it will thus both clean out the session's database record, and kill
        # all remaining pilots (none in our example).


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2016-2021, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import os

from .base import RMInfo, ResourceManager


# ------------------------------------------------------------------------------
#
class Torque(ResourceManager):

    # --------------------------------------------------------------------------
    #
    def _init_from_scratch(self, rm_info: RMInfo) -> RMInfo:

        nodefile = os.environ.get('PBS_NODEFILE')
        if not nodefile:
            raise RuntimeError('$PBS_NODEFILE not set')

        nodes = self._parse_nodefile(nodefile)

        if not rm_info.cores_per_node:
            rm_info.cores_per_node = self._get_cores_per_node(nodes)

        rm_info.node_list = self._get_node_list(nodes, rm_info)

        return rm_info


# ------------------------------------------------------------------------------





__copyright__ = 'Copyright 2014-2022, The RADICAL-Cybertools Team'
__license__   = 'MIT'

import copy
import os
import pprint
import stat
import time

import radical.utils       as ru

from ..   import utils     as rpu
from ..   import states    as rps
from ..   import constants as rpc
from ..   import TaskDescription
from ..db import DBSession

from .resource_manager import ResourceManager


# ------------------------------------------------------------------------------
#
class Agent_0(rpu.Worker):

    '''
    This is the main agent.  It starts sub-agents and watches them.  If any of
    the sub-agents die, it will shut down the other sub-agents and itself.

    This class inherits the rpu.Worker, so that it can use its communication
    bridges and callback mechanisms.  Specifically, it will pull the DB for
    new tasks to be executed and forwards them to the agent's component
    network (see `work()`).  It will also watch the DB for any commands to be
    forwarded (pilot termination, task cancellation, etc.), and will take care
    of heartbeat messages to be sent to the client module.  To do all this, it
    initializes a DB connection in `initialize()`.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self._uid     = 'agent.0'
        self._cfg     = cfg
        self._pid     = cfg.pid
        self._pmgr    = cfg.pmgr
        self._pwd     = cfg.pilot_sandbox
        self._session = session
        self._log     = ru.Logger(self._uid, ns='radical.pilot')

        self._starttime   = time.time()
        self._final_cause = None

        # this is the earliest point to sync bootstrap and agent profiles
        self._prof = ru.Profiler(ns='radical.pilot', name=self._uid)
        self._prof.prof('hostname', uid=cfg.pid, msg=ru.get_hostname())

        # run an inline registry service to share runtime config with other
        # agents and components
        reg_uid = 'radical.pilot.reg.%s' % self._uid
        self._reg_service = ru.zmq.Registry(uid=reg_uid)
        self._reg_service.start()

        # let all components know where to look for the registry
        self._cfg['reg_addr'] = self._reg_service.addr

        # connect to MongoDB for state push/pull
        self._connect_db()

        # configure ResourceManager before component startup, as components need
        # ResourceManager information for function (scheduler, executor)
        self._configure_rm()

        # ensure that app communication channels are visible to workload
        self._configure_app_comm()

        # expose heartbeat channel to sub-agents, bridges and components,
        # and start those
        self._cmgr = rpu.ComponentManager(self._cfg)
        self._cfg.heartbeat = self._cmgr.cfg.heartbeat

        self._cmgr.start_bridges()
        self._cmgr.start_components()

        # start any services if they are requested
        self._start_services()

        # create the sub-agent configs and start the sub agents
        self._write_sa_configs()
        self._start_sub_agents()   # TODO: move to cmgr?

        # at this point the session is up and connected, and it should have
        # brought up all communication bridges and components.  We are
        # ready to rumble!
        rpu.Worker.__init__(self, self._cfg, session)

        self.register_subscriber(rpc.CONTROL_PUBSUB, self._check_control)

        # run our own slow-paced heartbeat monitor to watch pmgr heartbeats
        # FIXME: we need to get pmgr freq
        freq = 60
        tint = freq / 3
        tout = freq * 10
        self._hb = ru.Heartbeat(uid=self._uid,
                                timeout=tout,
                                interval=tint,
                                beat_cb=self._hb_check,  # no own heartbeat(pmgr pulls)
                                term_cb=self._hb_term_cb,
                                log=self._log)
        self._hb.start()

        # register pmgr heartbeat
        self._log.info('hb init for %s', self._pmgr)
        self._hb.beat(uid=self._pmgr)


    # --------------------------------------------------------------------------
    #
    def _hb_check(self):

        self._log.debug('hb check')


    # --------------------------------------------------------------------------
    #
    def _hb_term_cb(self, msg=None):

        self._cmgr.close()
        self._log.warn('hb termination: %s', msg)

        return None


    # --------------------------------------------------------------------------
    #
    def _connect_db(self):

        # Check for the RADICAL_PILOT_DB_HOSTPORT env var, which will hold
        # the address of the tunnelized DB endpoint. If it exists, we
        # overrule the agent config with it.
        hostport = os.environ.get('RADICAL_PILOT_DB_HOSTPORT')
        if hostport:
            host, port = hostport.split(':', 1)
            dburl      = ru.Url(self._cfg.dburl)
            dburl.host = host
            dburl.port = port
            self._cfg.dburl = str(dburl)

        self._dbs = DBSession(sid=self._cfg.sid, dburl=self._cfg.dburl,
                              cfg=self._cfg, log=self._log)

    # --------------------------------------------------------------------------
    #
    def _configure_rm(self):

        # Create ResourceManager which will give us the set of agent_nodes to
        # use for sub-agent startup.  Add the remaining ResourceManager
        # information to the config, for the benefit of the scheduler).

        self._rm = ResourceManager.create(name=self._cfg.resource_manager,
                                          cfg=self._cfg, log=self._log,
                                          prof=self._prof)

        self._log.debug(pprint.pformat(self._rm.info))


    # --------------------------------------------------------------------------
    #
    def _configure_app_comm(self):

        # if the pilot description contains a request for application comm
        # channels, merge those into the agent config
        #
        # FIXME: this needs to start the app_comm bridges
        app_comm = self._cfg.get('app_comm')
        if app_comm:
            if isinstance(app_comm, list):
                app_comm = {ac: {'bulk_size': 0,
                                 'stall_hwm': 1,
                                 'log_level': 'error'} for ac in app_comm}
            for ac in app_comm:
                if ac in self._cfg['bridges']:
                    raise ValueError('reserved app_comm name %s' % ac)
                self._cfg['bridges'][ac] = app_comm[ac]


        # some of the bridge addresses also need to be exposed to the workload
        if app_comm:
            if 'task_environment' not in self._cfg:
                self._cfg['task_environment'] = dict()
            for ac in app_comm:
                if ac not in self._cfg['bridges']:
                    raise RuntimeError('missing app_comm %s' % ac)
                self._cfg['task_environment']['RP_%s_IN' % ac.upper()] = \
                        self._cfg['bridges'][ac]['addr_in']
                self._cfg['task_environment']['RP_%s_OUT' % ac.upper()] = \
                        self._cfg['bridges'][ac]['addr_out']


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        # registers the staging_input_queue as this is what we want to push
        # tasks to
        self.register_output(rps.AGENT_STAGING_INPUT_PENDING,
                             rpc.AGENT_STAGING_INPUT_QUEUE)

        # register the command callback which pulls the DB for commands
        self.register_timed_cb(self._agent_command_cb,
                               timer=self._cfg['db_poll_sleeptime'])

        # register idle callback to pull for tasks
        self.register_timed_cb(self._check_tasks_cb,
                               timer=self._cfg['db_poll_sleeptime'])

        # sub-agents are started, components are started, bridges are up: we are
        # ready to roll!  Update pilot state.
        pilot = {'type'             : 'pilot',
                 'uid'              : self._pid,
                 'state'            : rps.PMGR_ACTIVE,
                 'resource_details' : {
                     # 'lm_info'      : self._rm.lm_info.get('version_info'),
                     # 'lm_detail'    : self._rm.lm_info.get('lm_detail'),
                     'rm_info'      : self._rm.info},
                 '$set'             : ['resource_details']}
        self.advance(pilot, publish=True, push=False)


    # --------------------------------------------------------------------------
    #
    def work(self):

        # all work is done in the registered callbacks
        time.sleep(1)


    # --------------------------------------------------------------------------
    #
    def stage_output(self):

        if  os.path.isfile('./staging_output.txt'):

            if not os.path.isfile('./staging_output.tgz'):

                cmd = 'tar zcvf staging_output.tgz $(cat staging_output.txt)'
                out, err, ret = ru.sh_callout(cmd, shell=True)

                if ret:
                    self._log.debug('out: %s', out)
                    self._log.debug('err: %s', err)
                    self._log.error('output tarring failed: %s', cmd)


    # --------------------------------------------------------------------------
    #
    def finalize(self):

        # tar up output staging data
        self._log.debug('stage output parent')
        self.stage_output()

        # tear things down in reverse order
        self._hb.stop()
        self._cmgr.close()

        if self._rm:
            self._rm.stop()

        self._reg_service.stop()

        if   self._final_cause == 'timeout'  : state = rps.DONE
        elif self._final_cause == 'cancel'   : state = rps.CANCELED
        elif self._final_cause == 'sys.exit' : state = rps.CANCELED
        else                                 : state = rps.FAILED

        # NOTE: we do not push the final pilot state, as that is done by the
        #       bootstrapper *after* this pilot *actually* finished.
        with ru.ru_open('./killme.signal', 'w') as fout:
            fout.write('%s\n' % state)

        # we don't rely on the existence / viability of the update worker at
        # that point.
        self._log.debug('update db state: %s: %s', state, self._final_cause)
        self._log.info('rusage: %s', rpu.get_rusage())

        out, err, log = '', '', ''

        try   : out   = ru.ru_open('./agent.0.out', 'r').read(1024)
        except: pass
        try   : err   = ru.ru_open('./agent.0.err', 'r').read(1024)
        except: pass
        try   : log   = ru.ru_open('./agent.0.log', 'r').read(1024)
        except: pass

        ret = self._dbs._c.update({'type' : 'pilot',
                                   'uid'  : self._pid},
                                  {'$set' : {'stdout' : rpu.tail(out),
                                             'stderr' : rpu.tail(err),
                                             'logfile': rpu.tail(log),
                                             'state'  : state},
                                   '$push': {'states' : state}
                                  })
        self._log.debug('update ret: %s', ret)


    # --------------------------------------------------------------------
    #
    def _write_sa_configs(self):

        # we have all information needed by the subagents -- write the
        # sub-agent config files.

        # write deep-copies of the config for each sub-agent (sans from agent.0)
        for sa in self._cfg.get('agents', {}):

            assert (sa != 'agent.0'), 'expect subagent, not agent.0'

            # use our own config sans agents/components/bridges as a basis for
            # the sub-agent config.
            tmp_cfg = copy.deepcopy(self._cfg)
            tmp_cfg['agents']     = dict()
            tmp_cfg['components'] = dict()
            tmp_cfg['bridges']    = dict()

            # merge sub_agent layout into the config
            ru.dict_merge(tmp_cfg, self._cfg['agents'][sa], ru.OVERWRITE)

            tmp_cfg['uid']   = sa
            tmp_cfg['aid']   = sa
            tmp_cfg['owner'] = 'agent.0'

            ru.write_json(tmp_cfg, './%s.cfg' % sa)


    # --------------------------------------------------------------------------
    #
    def _start_services(self):
        '''
        If a `./services` file exist, reserve a compute node and run that file
        there as bash script.
        '''

        if not os.path.isfile('./services'):
            return

        # launch the `./services` script on the service node reserved by the RM.
        nodes = self._rm.info.service_node_list
        assert nodes

        bs_name = "%s/bootstrap_2.sh"     % self._pwd
        ls_name = "%s/services_launch.sh" % self._pwd
        ex_name = "%s/services_exec.sh"   % self._pwd

        service_task_uid = 'rp.services'
        service_task     = {
            'uid'               : service_task_uid,
            'task_sandbox_path' : self._pwd,
            'description'       : TaskDescription({
                'uid'           : service_task_uid,
                'ranks'         : 1,
                'cores_per_rank': self._rm.info.cores_per_node,
                'executable'    : '/bin/sh',
                'arguments'     : [bs_name, 'services']
            }).as_dict(),
            'slots': {'ranks'   : [{'node_name': nodes[0]['node_name'],
                                    'node_id'  : nodes[0]['node_id'],
                                    'core_map' : [[0]],
                                    'gpu_map'  : [],
                                    'lfs'      : 0,
                                    'mem'      : 0}]}
        }

        launcher = self._rm.find_launcher(service_task)
        if not launcher:
            raise RuntimeError('no launch method found for sub agent')

        # FIXME: set RP environment (as in Popen Executor)

        tmp  = '#!/bin/sh\n\n'
        tmp += 'export RP_PILOT_SANDBOX="%s"\n\n' % self._pwd
        cmds = launcher.get_launcher_env()
        for cmd in cmds:
            tmp += '%s || exit 1\n' % cmd
        cmds = launcher.get_launch_cmd(service_task, ex_name)
        tmp += '%s\nexit $?\n\n' % '\n'.join(cmds)

        with ru.ru_open(ls_name, 'w') as fout:
            fout.write(tmp)

        tmp  = '#!/bin/sh\n\n'
        tmp += '. ./env/service.env\n'
        tmp += '/bin/sh -l ./services\n\n'

        with ru.ru_open(ex_name, 'w') as fout:
            fout.write(tmp)


        # make sure scripts are executable
        st_l = os.stat(ls_name)
        st_e = os.stat(ex_name)
        os.chmod(ls_name, st_l.st_mode | stat.S_IEXEC)
        os.chmod(ex_name, st_e.st_mode | stat.S_IEXEC)

        # spawn the sub-agent
        cmdline = './%s' % ls_name

        self._log.info('create services: %s' % cmdline)
        ru.sh_callout_bg(cmdline, stdout='services.out', stderr='services.err')

        self._log.debug('services started done')


    # --------------------------------------------------------------------------
    #
    def _start_sub_agents(self):
        '''
        For the list of sub_agents, get a launch command and launch that
        agent instance on the respective node.  We pass it to the seconds
        bootstrap level, there is no need to pass the first one again.
        '''

        # FIXME: reroute to agent daemonizer

        if not self._cfg.get('agents'):
            return

        assert (len(self._rm.info.agent_node_list) >= len(self._cfg['agents']))

        self._log.debug('start_sub_agents')

        # store the current environment as the sub-agents will use the same
        ru.env_prep(os.environ, script_path='./env/agent.env')

        # the configs are written, and the sub-agents can be started.  To know
        # how to do that we create the agent launch method, have it creating
        # the respective command lines per agent instance, and run via
        # popen.
        #

        for idx, sa in enumerate(self._cfg['agents']):

            target = self._cfg['agents'][sa]['target']

            if target not in ['local', 'node']:

                raise ValueError('agent target unknown (%s)' % target)

            if target == 'local':

                # start agent locally
                cmdline = '/bin/sh -l %s/bootstrap_2.sh %s' % (self._pwd, sa)


            else:  # target == 'node':

                node = self._rm.info.agent_node_list[idx]
                # start agent remotely, use launch method
                # NOTE:  there is some implicit assumption that we can use
                #        the 'agent_node' string as 'agent_string:0' and
                #        obtain a well format slot...
                # FIXME: it is actually tricky to translate the agent_node
                #        into a viable 'slots' structure, as that is
                #        usually done by the schedulers.  So we leave that
                #        out for the moment, which will make this unable to
                #        work with a number of launch methods.  Can the
                #        offset computation be moved to the ResourceManager?
                bs_name       = '%s/bootstrap_2.sh' % (self._pwd)
                launch_script = '%s/%s.launch.sh'   % (self._pwd, sa)
                exec_script   = '%s/%s.exec.sh'     % (self._pwd, sa)

                agent_task = {
                    'uid'               : sa,
                    'task_sandbox_path' : self._pwd,
                    'description'       : TaskDescription({
                        'uid'           : sa,
                        'ranks'         : 1,
                        'cores_per_rank': self._rm.info.cores_per_node,
                        'executable'    : '/bin/sh',
                        'arguments'     : [bs_name, sa]
                    }).as_dict(),
                    'slots': {'ranks'   : [{'node_name': node['node_name'],
                                            'node_id'  : node['node_id'],
                                            'core_map' : [[0]],
                                            'gpu_map'  : [],
                                            'lfs'      : 0,
                                            'mem'      : 0}]}
                }

                # find a launcher to use
                launcher = self._rm.find_launcher(agent_task)
                if not launcher:
                    raise RuntimeError('no launch method found for sub agent')

                # FIXME: set RP environment (as in Popen Executor)

                tmp  = '#!/bin/sh\n\n'
                tmp += 'export RP_PILOT_SANDBOX="%s"\n\n' % self._pwd
                cmds = launcher.get_launcher_env()
                for cmd in cmds:
                    tmp += '%s || exit 1\n' % cmd

                cmds = launcher.get_launch_cmds(agent_task, exec_script)
                tmp += '%s\nexit $?\n\n' % cmds

                with ru.ru_open(launch_script, 'w') as fout:
                    fout.write(tmp)


                tmp  = '#!/bin/sh\n\n'
                tmp += '. ./env/agent.env\n'
                tmp += '/bin/sh -l ./bootstrap_2.sh %s\n\n' % sa

                with ru.ru_open(exec_script, 'w') as fout:
                    fout.write(tmp)

                # make sure scripts are executable
                st_l = os.stat(launch_script)
                st_e = os.stat(exec_script)
                os.chmod(launch_script, st_l.st_mode | stat.S_IEXEC)
                os.chmod(exec_script,   st_e.st_mode | stat.S_IEXEC)

                # spawn the sub-agent
                cmdline = launch_script

            self._log.info ('create sub-agent %s: %s' % (sa, cmdline))
            ru.sh_callout_bg(cmdline, stdout='%s.out' % sa,
                                      stderr='%s.err' % sa)

            # FIXME: register heartbeats?

        self._log.debug('start_sub_agents done')


    # --------------------------------------------------------------------------
    #
    def _agent_command_cb(self):

        if not self._check_commands(): return False
        if not self._check_rpc     (): return False
        if not self._check_state   (): return False

        return True


    # --------------------------------------------------------------------------
    #
    def _check_commands(self):

        # Check if there's a command waiting
        # FIXME: this pull should be done by the update worker, and commands
        #        should then be communicated over the command pubsub
        # FIXME: commands go to pmgr, tmgr, session docs
        # FIXME: check if pull/wipe are atomic
        # FIXME: long runnign commands can time out on hb
        retdoc = self._dbs._c.find_and_modify(
                    query ={'uid' : self._pid},
                    fields=['cmds'],                    # get  new commands
                    update={'$set': {'cmds': list()}})  # wipe old commands

        if not retdoc:
            return True

        for spec in retdoc.get('cmds', []):

            cmd = spec['cmd']
            arg = spec['arg']

            self._log.debug('pilot command: %s: %s', cmd, arg)
            self._prof.prof('cmd', msg="%s : %s" %  (cmd, arg), uid=self._pid)

            if cmd == 'heartbeat' and arg['pmgr'] == self._pmgr:
                self._hb.beat(uid=self._pmgr)

            elif cmd == 'cancel_pilot':
                self._log.info('cancel pilot cmd')
                self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'terminate',
                                                  'arg' : None})
                self._final_cause = 'cancel'
                return False  # we are done

            elif cmd == 'cancel_tasks':
                self._log.info('cancel_tasks cmd')
                self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'cancel_tasks',
                                                  'arg' : arg})
            else:
                self._log.warn('could not interpret cmd "%s" - ignore', cmd)

        return True


    # --------------------------------------------------------------------------
    #
    def _check_rpc(self):
        '''
        check if the DB has any RPC request for this pilot.  If so, then forward
        that request as `rpc_req` command on the CONTROL channel, and listen for
        an `rpc_res` command on the same channel, for the same rpc id.  Once
        that response is received (from whatever component handled that
        command), send the response back to the databse for the callee to pick
        up.
        '''

        # FIXME: implement a timeout, and/or a registry of rpc clients

        retdoc = self._dbs._c.find_and_modify(
                    query ={'uid' : self._pid},
                    fields=['rpc_req'],
                    update={'$set': {'rpc_req': None}})

        if not retdoc:
            # no rpc request found
            return True

        rpc_req = retdoc.get('rpc_req')
        if rpc_req is None:
            # document has no rpc request
            return True

        self._log.debug('rpc req: %s', rpc_req)

        # RPCs are synchronous right now - we send the RPC on the command
        # channel, hope that some component picks it up and replies, and then
        # return that reply.  The reply is received via a temporary callback
        # defined here, which will receive all CONTROL messages until the right
        # rpc response comes along.
        def rpc_cb(topic, msg):

            rpc_id  = rpc_req['uid']

            cmd     = msg['cmd']
            rpc_res = msg['arg']

            if cmd != 'rpc_res':
                # not an rpc responese
                return True

            if rpc_res['uid'] != rpc_id:
                # not the right rpc response
                return True

            # send the response to the DB
            self._dbs._c.update({'type'  : 'pilot',
                                 'uid'   : self._pid},
                                {'$set'  : {'rpc_res': rpc_res}})

            # work is done - unregister this temporary cb (rpc_cb)
            return False


        self.register_subscriber(rpc.CONTROL_PUBSUB, rpc_cb)

        # ready to receive and proxy rpc response -- forward rpc request on
        # control channel
        self.publish(rpc.CONTROL_PUBSUB, {'cmd' : 'rpc_req',
                                          'arg' :  rpc_req})

        return True  # keeb cb registered (self._check_rpc)


    # --------------------------------------------------------------------------
    #
    def _check_control(self, _, msg):
        '''
        Check for commands on the control pubsub, mainly waiting for RPC
        requests to handle.  We handle two types of RPC requests: `hello` for
        testing, and `prepare_env` for environment preparation requests.
        '''

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd != 'rpc_req':
            # not an rpc request
            return True

        req = arg['rpc']
        if req not in ['hello', 'prepare_env']:
            # we don't handle that request
            return True

        rpc_res = {'uid': arg['uid']}
        try:
            if req == 'hello'   :
                out = 'hello %s' % ' '.join(arg['arg'])

            elif req == 'prepare_env':
                env_name = arg['arg']['env_name']
                env_spec = arg['arg']['env_spec']
                out      = self._prepare_env(env_name, env_spec)

            else:
                # unknown command
                return True

            # request succeeded - respond with return value
            rpc_res['err'] = None
            rpc_res['out'] = out
            rpc_res['ret'] = 0

        except Exception as e:
            # request failed for some reason - indicate error
            rpc_res['err'] = repr(e)
            rpc_res['out'] = None
            rpc_res['ret'] = 1
            self._log.exception('control cmd failed')

        # publish the response (success or failure)
        self.publish(rpc.CONTROL_PUBSUB, {'cmd': 'rpc_res',
                                          'arg':  rpc_res})
        return True


    # --------------------------------------------------------------------------
    #
    def _check_state(self):

        # Make sure that we haven't exceeded the runtime - otherwise terminate.
        if self._cfg.runtime:

            if time.time() >= self._starttime +  (int(self._cfg.runtime) * 60):

                self._log.info('runtime limit (%ss).', self._cfg.runtime * 60)
                self._final_cause = 'timeout'
                self.stop()
                return False  # we are done

        return True


    # --------------------------------------------------------------------------
    #
    def _check_tasks_cb(self):

        # Check for tasks waiting for input staging and log pull.
        #
        # FIXME: Unfortunately, 'find_and_modify' is not bulkable, so we have
        #        to use 'find'.  To avoid finding the same tasks over and over
        #        again, we update the 'control' field *before* running the next
        #        find -- so we do it right here.
        #        This also blocks us from using multiple ingest threads, or from
        #        doing late binding by task pull :/
        task_cursor = self._dbs._c.find({'type'    : 'task',
                                         'pilot'   : self._pid,
                                         'control' : 'agent_pending'})
        if not task_cursor.count():
            self._log.info('tasks pulled:    0')
            return True

        # update the tasks to avoid pulling them again next time.
        task_list = list(task_cursor)
        task_uids = [task['uid'] for task in task_list]

        self._dbs._c.update({'type'  : 'task',
                             'uid'   : {'$in'     : task_uids}},
                            {'$set'  : {'control' : 'agent'}},
                            multi=True)

        self._log.info("tasks pulled: %4d", len(task_list))
        self._prof.prof('get', msg='bulk: %d' % len(task_list), uid=self._pid)

        for task in task_list:

            # make sure the tasks obtain env settings (if needed)
            if 'task_environment' in self._cfg:
                if not task['description'].get('environment'):
                    task['description']['environment'] = dict()
                for k,v in self._cfg['task_environment'].items():
                    task['description']['environment'][k] = v

            # we need to make sure to have the correct state:
            task['state'] = rps._task_state_collapse(task['states'])
            self._prof.prof('get', uid=task['uid'])

            # FIXME: raise or fail task!
            if task['state'] != rps.AGENT_STAGING_INPUT_PENDING:
                self._log.error('invalid state: %s', (pprint.pformat(task)))

            task['control'] = 'agent'

        # now we really own the CUs, and can start working on them (ie. push
        # them into the pipeline).  We don't publish nor profile as advance,
        # since that happened already on the module side when the state was set.
        self.advance(task_list, publish=False, push=True)

        return True


    # --------------------------------------------------------------------------
    #
    def _prepare_env(self, env_name, env_spec):

        self._log.debug('env_spec: %s', env_spec)

        etype = env_spec.get('type', 'venv')
        evers = env_spec.get('version')
        path  = env_spec.get('path')
        emods = env_spec.get('setup')    or []
        pre   = env_spec.get('pre_exec') or []
        out   = None

        pre_exec = '-P ". env/bs0_pre_0.sh" '
        for cmd in pre:
            pre_exec += '-P "%s" ' % cmd

        if emods: mods = '-m "%s"' % ','.join(emods)
        else    : mods = ''

      # assert etype == 'virtualenv'
      # assert evers

        # only create a new VE if path is not set or if it does not exist
        if path:
            path = path.rstrip('/')

        ve_local_path = '%s/env/rp_named_env.%s' % (self._pwd, env_name)
        if path: ve_path = path
        else   : ve_path = ve_local_path

        if evers:
            evers = '-v %s' % evers
        else:
            evers = ''

        rp_cse = ru.which('radical-pilot-create-static-ve')
        ve_cmd = '/bin/bash %s -d -p %s -t %s %s %s %s > env.log 2>&1' \
               % (rp_cse, ve_path, etype, evers, mods, pre_exec)

        # FIXME: we should export all sandboxes etc. to the prep_env.
        os.environ['RP_RESOURCE_SANDBOX'] = '../../'

        self._log.debug('env cmd: %s', ve_cmd)
        out, err, ret = ru.sh_callout(ve_cmd, shell=True)
        self._log.debug('    out: %s', out)
        self._log.debug('    err: %s', err)

        if ret:
            raise RuntimeError('prepare_env failed: \n%s\n%s\n' % (out, err))

        # if the ve lives outside of the pilot sandbox, link it
        if path:
            os.symlink(path,          ve_local_path)
            os.symlink(path + '.env', ve_local_path + '.env')

        self._log.debug('ve_path: %s', ve_path)

        # prepare the env to be loaded in task exec scripts
        sh_path = '%s/env/rp_named_env.%s.sh' % (self._pwd, env_name)
        with ru.ru_open(sh_path, 'w') as fout:
            fout.write('\n. %s/bin/activate\n\n' % ve_path)

        # publish the venv creation to the scheduler
        self.publish(rpc.CONTROL_PUBSUB, {'cmd': 'register_named_env',
                                          'arg': {'env_name': env_name}})
        return out


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

__copyright__ = 'Copyright 2013-2014, http://radical.rutgers.edu'
__license__   = 'MIT'


import os
import radical.pilot as rp

n_master =  1
n_worker =  16


# ------------------------------------------------------------------------------
#
if __name__ == '__main__':

    session = rp.Session()

    try:
        pmgr    = rp.PilotManager(session=session)
        pd_init = {'resource'      : 'local.localhost',
                   'runtime'       : 60,
                   'exit_on_error' : True,
                   'cores'         : n_master * 1 + n_worker * 1,
                   'app_comm'      : ['app_pubsub',
                                      'work_queue',
                                      'result_queue']
                  }
        pdesc = rp.PilotDescription(pd_init)
        pilot = pmgr.submit_pilots(pdesc)
        tmgr  = rp.TaskManager(session=session)
        tmgr.add_pilots(pilot)

        pwd  = os.path.dirname(os.path.abspath(__file__))
        TD  = rp.TaskDescription
        tds = list()

        for i in range(n_master):
            tds.append(TD({'executable': '%s/rp_app_master.py' % pwd,
                             'arguments' : [n_worker]}))
        for i in range(n_worker):
            tds.append(TD({'executable': '%s/rp_app_worker.py' % pwd}))

        tmgr.submit_tasks(tds)
        tmgr.wait_tasks()

    finally:
        session.close(download=True)


# ------------------------------------------------------------------------------





import os
import sys
import time

import radical.utils     as ru

from .. import constants as rpc


# ------------------------------------------------------------------------------
#
class Worker(object):
    '''
    Implement the Raptor protocol for dispatching multiple Tasks on persistent
    resources.
    '''

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg=None, register=True, session=None):

        self._session = session

        if cfg is None:
            cfg = dict()

        if isinstance(cfg, str): self._cfg = ru.Config(cfg=ru.read_json(cfg))
        else                   : self._cfg = ru.Config(cfg=cfg)

        self._sbox = os.environ['RP_TASK_SANDBOX']
        self._uid  = os.environ['RP_TASK_ID']

        self._log  = ru.Logger(name=self._uid,   ns='radical.pilot.worker',
                               level='DEBUG', targets=['.'], path=self._sbox)
        self._prof = ru.Profiler(name=self._uid, ns='radical.pilot.worker',
                                 path=self._sbox)

        if register:

            psbox = os.environ['RP_PILOT_SANDBOX']
            ctrl_cfg = ru.read_json('%s/%s.cfg' % (psbox, rpc.CONTROL_PUBSUB))

            self._control_sub = ru.zmq.Subscriber(rpc.CONTROL_PUBSUB,
                                                  url=ctrl_cfg['sub'],
                                                  log=self._log,
                                                  prof=self._prof,
                                                  cb=self._control_cb)

            self._ctrl_pub = ru.zmq.Publisher(rpc.CONTROL_PUBSUB,
                                              url=ctrl_cfg['pub'],
                                              log=self._log,
                                              prof=self._prof)
            # let ZMQ settle
            time.sleep(1)

            # `info` is a placeholder for any additional meta data
            self._ctrl_pub.put(rpc.CONTROL_PUBSUB, {'cmd': 'worker_register',
                                                    'arg': {'uid' : self._uid,
                                                            'info': {}}})

          # # FIXME: we never unregister on termination
          # self._ctrl_pub.put(rpc.CONTROL_PUBSUB, {'cmd': 'worker_unregister',
          #                                         'arg': {'uid' : self._uid}})


    # --------------------------------------------------------------------------
    #
    @staticmethod
    def run(fpath, cname, cfg):

        # load worker class from fname if that is a valid string
        wclass = None

        # Create the worker class and run it's work loop.
        if fpath:
            wclass = ru.load_class(fpath, cname, Worker)

        else:
            # import all known workers into the local name space so that
            # `get_type` has a chance to find them

            from .worker_default import DefaultWorker  # pylint: disable=W0611 # noqa
            from .worker_mpi     import MPIWorker      # pylint: disable=W0611 # noqa

            wclass = ru.get_type(cname)

        if not wclass:
            raise RuntimeError('no worker [%s] [%s]' % (cname, fpath))

        worker = wclass(cfg)
        worker.start()
        worker.join()


    # --------------------------------------------------------------------------
    #
    def _control_cb(self, topic, msg):

        if msg['cmd'] == 'terminate':
            self.stop()
            self.join()
            sys.exit()

        elif msg['cmd'] == 'worker_terminate':
            self._log.debug('worker_terminate signal')
            if msg['arg']['uid'] == self._uid:
                self.stop()
                self.join()
                sys.exit()


    # --------------------------------------------------------------------------
    #
    def get_master(self):
        '''
        The worker can submit tasks back to the master - this method will
        return a small shim class to provide that capability
        '''

        # ----------------------------------------------------------------------
        class Master(object):

            def __init__(self, addr):
                self._task_service_ep = ru.zmq.Client(url=addr)

            def run_task(self, td):
                return self._task_service_ep.request('run_task', td)
        # ----------------------------------------------------------------------

        return Master(self._cfg.ts_addr)


    # --------------------------------------------------------------------------
    #
    def start(self):

        raise NotImplementedError('`start()` must be implemented by child class')


    # --------------------------------------------------------------------------
    #
    def stop(self):

        raise NotImplementedError('`run()` must be implemented by child class')


    # --------------------------------------------------------------------------
    #
    def join(self):

        raise NotImplementedError('`join()` must be implemented by child class')


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"


import os

import radical.utils as ru

from ... import utils     as rpu
from ... import states    as rps
from ... import constants as rpc


# ------------------------------------------------------------------------------
# 'enum' for RPs's tmgr scheduler types
SCHEDULER_ROUND_ROBIN  = "round_robin"
SCHEDULER_BACKFILLING  = "backfilling"

# default:
SCHEDULER_DEFAULT      = SCHEDULER_ROUND_ROBIN

# internally used enums for pilot roles
ROLE    = '_scheduler_role'
ADDED   = 'added'
REMOVED = 'removed'
FAILED  = 'failed'


# ------------------------------------------------------------------------------
#
class TMGRSchedulingComponent(rpu.Component):

    # FIXME: clarify what can be overloaded by Scheduler classes

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        self._uid = ru.generate_id(cfg['owner'] + '.scheduling.%(counter)s',
                                   ru.ID_CUSTOM)

        rpu.Component.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def initialize(self):

        self._tmgr = self._cfg.owner

        self._early        = dict()      # early-bound tasks, pid-sorted
        self._pilots       = dict()      # dict of known pilots
        self._pilots_lock  = ru.RLock()  # lock on the above dict
        self._tasks        = dict()      # dict of scheduled task IDs
        self._tasks_lock   = ru.RLock()  # lock on the above dict
        self._waiting      = dict()      # dict for tasks waiting on deps
        self._waiting_lock = dict()      # lock on the above dict

        # configure the scheduler instance
        self._configure()

        self.register_input(rps.TMGR_SCHEDULING_PENDING,
                            rpc.TMGR_SCHEDULING_QUEUE, self.work)

        self.register_output(rps.TMGR_STAGING_INPUT_PENDING,
                             rpc.TMGR_STAGING_INPUT_QUEUE)

        # Some schedulers care about states (of pilots and/or tasks), some
        # don't.  Either way, we here subscribe to state updates.
        self.register_subscriber(rpc.STATE_PUBSUB, self._base_state_cb)

        # Schedulers use that command channel to get information about
        # pilots being added or removed.
        self.register_subscriber(rpc.CONTROL_PUBSUB, self._base_command_cb)

        # cache the local client sandbox to avoid repeated os calls
        self._client_sandbox = os.getcwd()


    # --------------------------------------------------------------------------
    #
    # This class-method creates the appropriate sub-class for the Scheduler.
    #
    @classmethod
    def create(cls, cfg, session):

        # Make sure that we are the base-class!
        if cls != TMGRSchedulingComponent:
            raise TypeError("Scheduler Factory only available to base class!")

        name = cfg['scheduler']

        from .round_robin  import RoundRobin
        from .backfilling  import Backfilling

        impl = {
            SCHEDULER_ROUND_ROBIN : RoundRobin,
            SCHEDULER_BACKFILLING : Backfilling
        }

        if name not in impl:
            raise ValueError('Scheduler %s unknown' % name)

        return impl[name](cfg, session)


    # --------------------------------------------------------------------------
    #
    def _base_state_cb(self, topic, msg):

        # the base class will keep track of pilot state changes and updates
        # self._pilots accordingly.  Task state changes will also be collected,
        # but not stored - if a scheduler needs to keep track of task state
        # changes, it needs to overload `update_tasks()`.

        cmd = msg.get('cmd')
        arg = msg.get('arg')

        self._log.info('scheduler state_cb: %s', cmd)

        # FIXME: get cmd string consistent throughout the code
        if cmd not in ['update', 'state_update']:
            self._log.debug('ignore cmd %s', cmd)
            return True

        if not isinstance(arg, list): things = [arg]
        else                        : things =  arg

        pilots = [t for t in things if t['type'] == 'pilot']
        tasks  = [t for t in things if t['type'] == 'task' ]

        self._log.debug('update pilots %s', [p['uid'] for p in pilots])
        self._log.debug('update tasks  %s', [u['uid'] for u in tasks])

        self._update_pilot_states(pilots)
        self._update_task_states(tasks)

        return True


    # --------------------------------------------------------------------------
    #
    def _update_pilot_states(self, pilots):

        self._log.debug('update pilot states for %s', [p['uid'] for p in pilots])

        if not pilots:
            return

        to_update = list()

        with self._pilots_lock:

            for pilot in pilots:

                pid = pilot['uid']

                if pid not in self._pilots:
                    self._pilots[pid] = {'role'  : None,
                                         'state' : None,
                                         'pilot' : None,
                                         'info'  : dict()  # scheduler private info
                                         }

                target  = pilot['state']
                current = self._pilots[pid]['state']

                # enforce state model order
                target, passed = rps._pilot_state_progress(pid, current, target)

                if current != target:
                    to_update.append(pid)
                    self._pilots[pid]['state'] = target
                    self._log.debug('update pilot state: %s -> %s', current, passed)

        if to_update:
            self.update_pilots(to_update)
        self._log.debug('updated  : %s', to_update)


    # --------------------------------------------------------------------------
    #
    def _update_task_states(self, tasks):

        self.update_tasks(tasks)


    # --------------------------------------------------------------------------
    #
    def update_tasks(self, tasks):
        '''
        any scheduler that cares about task state changes should implement this
        method to keep track of those
        '''

        pass


    # --------------------------------------------------------------------------
    #
    def _base_command_cb(self, topic, msg):

        # we'll wait for commands from the tmgr, to learn about pilots we can
        # use or we should stop using. We also track task cancelation, as all
        # components do.
        #
        # make sure command is for *this* scheduler by matching the tmgr uid.

        cmd = msg['cmd']

        self._log.debug('got cmd %s', cmd)

        if cmd not in ['add_pilots', 'remove_pilots', 'cancel_tasks']:
            return True

        arg   = msg['arg']
        tmgr  = arg['tmgr']

        self._log.info('scheduler command: %s: %s' % (cmd, arg))

        if tmgr and tmgr != self._tmgr:
            # this is not the command we are looking for
            return True


        if cmd == 'add_pilots':

            pilots = arg['pilots']

            with self._pilots_lock:

                for pilot in pilots:

                    pid = pilot['uid']

                    if pid in self._pilots:
                        if self._pilots[pid]['role'] == ADDED:
                            raise ValueError('pilot already added (%s)' % pid)
                    else:
                        self._pilots[pid] = {'role'  : None,
                                             'state' : None,
                                             'pilot' : None,
                                             'info'  : dict()
                                            }

                    self._pilots[pid]['role']  = ADDED
                    self._pilots[pid]['pilot'] = pilot
                    self._log.debug('added pilot: %s', self._pilots[pid])

                self._update_pilot_states(pilots)

                for pilot in pilots:

                    pid = pilot['uid']

                    # if we have any early_bound tasks waiting for this pilots,
                    # advance them now
                    early_tasks = self._early.get(pid)
                    if early_tasks:

                        for task in early_tasks:
                            self._assign_pilot(task, pilot)

                        self.advance(early_tasks, rps.TMGR_STAGING_INPUT_PENDING,
                                     publish=True, push=True)

            # let the scheduler know
            self.add_pilots([pilot['uid'] for pilot in pilots])


        elif cmd == 'remove_pilots':

            pids = arg['pids']

            with self._pilots_lock:

                for pid in pids:

                    if pid not in self._pilots:
                        raise ValueError('pilot not added (%s)' % pid)

                    if self._pilots[pid]['role'] != ADDED:
                        raise ValueError('pilot not added (%s)' % pid)

                    self._pilots[pid]['role'] = REMOVED
                    self._log.debug('removed pilot: %s', self._pilots[pid])

            # let the scheduler know
            self.remove_pilots(pids)


        elif cmd == 'cancel_tasks':

            uids = arg['uids']

            # find the pilots handling these tasks and forward the cancellation
            # request
            to_cancel = dict()

            with self._tasks_lock:
                for pid in self._tasks:
                    for uid in uids:
                        if uid in self._tasks[pid]:
                            if pid not in to_cancel:
                                to_cancel[pid] = list()
                            to_cancel[pid].append(uid)

            dbs = self._session._dbs

            if not dbs:
                # too late, already closing down
                return True

            for pid in to_cancel:
                dbs.pilot_command(cmd='cancel_tasks',
                                  arg={'uids' : to_cancel[pid]},
                                  pids=pid)

        return True


    # --------------------------------------------------------------------------
    #
    def _configure(self):
        raise NotImplementedError("_configure() missing for '%s'" % self.uid)


    # --------------------------------------------------------------------------
    #
    def _assign_pilot(self, task, pilot):
        '''
        assign a task to a pilot.
        This is also a good opportunity to determine the task sandbox(es).
        '''

        pid = pilot['uid']
        uid = task['uid']

        self._log.debug('assign %s to %s', uid, pid)

        task['pilot'            ] = pid
        task['client_sandbox'   ] = str(self._session._get_client_sandbox())
        task['endpoint_fs'      ] = str(self._session._get_endpoint_fs(pilot))
        task['resource_sandbox' ] = str(self._session._get_resource_sandbox(pilot))
        task['session_sandbox'  ] = str(self._session._get_session_sandbox(pilot))
        task['pilot_sandbox'    ] = str(self._session._get_pilot_sandbox(pilot))
        task['task_sandbox'     ] = str(self._session._get_task_sandbox(task, pilot))
        task['task_sandbox_path'] = ru.Url(task['task_sandbox']).path

        with self._tasks_lock:
            if pid not in self._tasks:
                self._tasks[pid] = list()
            self._tasks[pid].append(uid)


    # --------------------------------------------------------------------------
    #
    def add_pilots(self, pids):
        raise NotImplementedError("add_pilots() missing for '%s'" % self.uid)


    # --------------------------------------------------------------------------
    #
    def remove_pilots(self, pids):
        raise NotImplementedError("remove_pilots() missing for '%s'" % self.uid)


    # --------------------------------------------------------------------------
    #
    def update_pilots(self, pids):
        raise NotImplementedError("update_pilots() missing for '%s'" % self.uid)


    # --------------------------------------------------------------------------
    #
    def work(self, tasks):
        '''
        We get a number of tasks, and filter out those which are already bound
        to a pilot.  Those will get advanced to TMGR_STAGING_INPUT_PENDING
        straight away.  All other tasks are passed on to `self._work()`, which
        is the scheduling routine as implemented by the deriving scheduler
        classes.

        Note that the filter needs to know any pilot the tasks are early-bound
        to, as it needs to obtain the task's sandbox information to prepare for
        the follow-up staging ops.  If the respective pilot is not yet known,
        the tasks are put into a early-bound-wait list, which is checked and
        emptied as pilots get registered.
        '''

        if not isinstance(tasks, list):
            tasks = [tasks]

        # some task may have staging directives which reference sandboxes of
        # other tasks.  Those directives can only be expanded with actual
        # physical path's once both tasks are known to the scheduler, so we
        # check this here and (a) let all tasks wait until the references are
        # resolved, and (b) check if the tasks resolve any references
        #
        # The `waiting` data structure has the following format:
        #
        #   {
        #     'waiting': {
        #         <uid_a> : {
        #             'task': <task>,
        #             'deps': [<uid_1>, <uid_2>, ...]
        #         },
        #         ...
        #       },
        #     'deps': {
        #       <uid_1> : [<uid_a>, <uid_b>, ...]
        #       ...
        #   }
        #
        # for each incoming tasks <uid_1>, we check in `deps` if depending tasks
        # are known, then remove <uid_1> from the global `deps` dict and also
        # from the `deps` list of each of those waiting tasks.  If any of those
        # waiting tasks then ends up with an empty `deps` list, then that task
        # will not be waiting anymore and can be scheduled.  We mark both
        # participating tasks so that the scheduler can ensure they end up on
        # the same pilot.
        #
        # NOTE: cross-pilot data dependencies are not yet supported
        #
        # The task staging directives are expected to be expanded to their
        # dictionary format already, and will check for `src` or `tgt` URLs with
        # a `sandbox://` schema, where the `host` element can reference
        #
        #   client:   the client application pwd
        #   resource: the target resource sandbox
        #   pilot:    the target pilot's sandbox
        #   <uid>:    the sandbox of the respective task
        #
        # This implies that `client`, `resource` and `pilot` are reserved names
        # for task IDs, and that tasks which use invalid / non-existing IDs in
        # sandbox references will never be eligible for scheduling.

        self.advance(tasks, rps.TMGR_SCHEDULING, publish=True, push=False)

        to_schedule = list()

        with self._pilots_lock:

            for task in tasks:

                uid = task['uid']
                pid = task.get('pilot')

                if pid:
                    # this task is bound already (it is early-bound), so we
                    # don't need to pass it to the actual scheduling algorithm

                    # check if we know about the pilot, so that we can advance
                    # the task to data staging
                    pilot = self._pilots.get(pid, {}).get('pilot')
                    if pilot:
                        self._assign_pilot(task, pilot)
                        self.advance(task, rps.TMGR_STAGING_INPUT_PENDING,
                                     publish=True, push=True)

                    else:
                        # otherwise keep in `self._early` until we learn about
                        # the pilot
                        self._log.warn('got task %s for unknown pilot %s', uid, pid)
                        if pid not in self._early:
                            self._early[pid] = list()
                        self._early[pid].append(task)

                else:
                    to_schedule.append(task)

        self._log.debug('to_schedule: %d', len(to_schedule))
        self._work(to_schedule)


    # --------------------------------------------------------------------------
    #
    def _work(self, tasks):

        raise NotImplementedError("work() missing for '%s'" % self.uid)


# ------------------------------------------------------------------------------






__copyright__ = 'Copyright 2013-2016, http://radical.rutgers.edu'
__license__   = 'MIT'


import os
import stat
import time
import queue
import atexit
import pprint
import signal
import threading as mt
import subprocess

import radical.utils as ru

from ...  import states    as rps
from ...  import constants as rpc

from .base import AgentExecutingComponent


# ------------------------------------------------------------------------------
# ensure tasks are killed on termination
_pids = list()


# pylint: disable=unused-argument
def _kill(*args, **kwargs):

    for pid in _pids:

        # skip test mocks
        if not isinstance(pid, int):
            continue

        try   : os.killpg(pid, signal.SIGTERM)
        except: pass


atexit.register(_kill)
signal.signal(signal.SIGTERM, _kill)
signal.signal(signal.SIGINT,  _kill)
# ------------------------------------------------------------------------------


# ------------------------------------------------------------------------------
#
class Popen(AgentExecutingComponent):

    _header    = '#!/bin/sh\n'
    _separator = '\n# ' + '-' * 78 + '\n'

    # --------------------------------------------------------------------------
    #
    def initialize(self):

      # self._log.debug('popen initialize start')
        AgentExecutingComponent.initialize(self)

        self._cancel_lock     = mt.RLock()
        self._tasks_to_cancel = list()
        self._tasks_to_watch  = list()
        self._watch_queue     = queue.Queue()

        self._pid = self._cfg['pid']

        # run watcher thread
        self._watcher = mt.Thread(target=self._watch)
      # self._watcher.daemon = True
        self._watcher.start()

      # self._log.debug('popen initialize stop')

    # --------------------------------------------------------------------------
    #
    def command_cb(self, topic, msg):

        self._log.info('command_cb [%s]: %s', topic, msg)

        cmd = msg['cmd']
        arg = msg['arg']

        if cmd == 'cancel_tasks':

            self._log.info('cancel_tasks command (%s)' % arg)
            with self._cancel_lock:
                self._tasks_to_cancel.extend(arg['uids'])

        return True

    # --------------------------------------------------------------------------
    #
    def work(self, tasks):

        self.advance(tasks, rps.AGENT_EXECUTING, publish=True, push=False)

        for task in tasks:

            try:
                self._prof.prof('task_start', uid=task['uid'])
                self._handle_task(task)

            except Exception:
                # append the startup error to the tasks stderr.  This is
                # not completely correct (as this text is not produced
                # by the task), but it seems the most intuitive way to
                # communicate that error to the application/user.
                self._log.exception("error running Task")
                if task['stderr'] is None:
                    task['stderr'] = ''
                task['stderr'] += '\nPilot cannot start task:\n'
                task['stderr'] += '\n'.join(ru.get_exception_trace())

                # can't rely on the executor base to free the task resources
                self._prof.prof('unschedule_start', uid=task['uid'])
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

                task['control'] = 'tmgr_pending'
                task['$all']    = True
                self.advance(task, rps.FAILED, publish=True, push=False)

            finally:
                self._prof.prof('task_stop', uid=task['uid'])


    # --------------------------------------------------------------------------
    #
    def _handle_task(self, task):

        # before we start handling the task, check if it should run in a named
        # env.  If so, inject the activation of that env in the task's pre_exec
        # directives.
        tid  = task['uid']
        td   = task['description']
        sbox = task['task_sandbox_path']

        # prepare stdout/stderr
        task['stdout'] = ''
        task['stderr'] = ''

        stdout_file    = td.get('stdout') or '%s.out' % (tid)
        stderr_file    = td.get('stderr') or '%s.err' % (tid)

        if stdout_file[0] != '/':
            task['stdout_file']       = '%s/%s' % (sbox, stdout_file)
            task['stdout_file_short'] = '$RP_TASK_SANDBOX/%s' % (stdout_file)
        else:
            task['stdout_file']       = stdout_file
            task['stdout_file_short'] = stdout_file

        if stderr_file[0] != '/':
            task['stderr_file']       = '%s/%s' % (sbox, stderr_file)
            task['stderr_file_short'] = '$RP_TASK_SANDBOX/%s' % (stderr_file)
        else:
            task['stderr_file']       = stderr_file
            task['stderr_file_short'] = stderr_file

        # create two shell scripts: a launcher script (task.launch.sh) which
        # sets the launcher environment, performs pre_launch commands, and then
        # launches the second script which executes the task.
        #
        # The second script (`task.exec.sh`) is instantiated once per task rank.
        # It first resets the environment created by the launcher, then prepares
        # the environment for the tasks.  Next it runs the `pre_exec` directives
        # for all ranks, then the individual `pre_rank` directives are executed,
        # and then, after all ranks are synchronized, finally the task ranks
        # begin to run.
        #
        # The scripts thus show the following structure:
        #
        # Launcher Script (`task.000000.launch.sh`):
        # ----------------------------------------------------------------------
        # #!/bin/sh
        #
        # # `pre_launch` commands
        # date > data.input
        #
        # # launcher specific environment setup
        # module load openmpi
        #
        # # launch the task script
        # mpirun -n 4 ./task.000000.exec.sh
        # ----------------------------------------------------------------------
        #
        # Task Script (`task.000000.exec.sh`)
        # ----------------------------------------------------------------------
        # #!/bin/sh
        #
        # # clean launch environment
        # module unload mpi
        #
        # # task environment setup (`pre_exec`)
        # module load gromacs
        #
        # # rank specific setup (`pre_rank`)
        # touch task.000000.ranks
        # if test "$MPI_RANK" = 0; then
        #   export CUDA_VISIBLE_DEVICES=0
        #   export OENMP_NUM_THREADS=2
        #   export RANK_0_VAR=foo
        #   echo 0 >> task.000000.ranks
        # fi
        #
        # if test "$MPI_RANK" = 1; then
        #   export CUDA_VISIBLE_DEVICES=1
        #   export OENMP_NUM_THREADS=4
        #   export RANK_1_VAR=bar
        #   echo 1 >> task.000000.ranks
        # fi
        #
        # # synchronize ranks
        # while $(cat task.000000.ranks | wc -l) != $MPI_RANKS; do
        #   sleep 1
        # done
        #
        # # run the task
        # mdrun -i ... -o ... -foo ... 1> task.000000.$MPI_RANK.out \
        #                              2> task.000000.$MPI_RANK.err
        #
        # # now do the very same stuff for the `post_rank` and `post_exec`
        # # directives
        # ...
        #
        # ----------------------------------------------------------------------
        #
        # NOTE: MongoDB only accepts string keys, and thus the rank IDs in
        #       pre_rank and post_rank dictionaries are rendered as strings.
        #       This should be changed to more intuitive integers once MongoDB
        #       is phased out.
        #
        # prep stdout/err so that we can append w/o checking for None
        task['stdout'] = ''
        task['stderr'] = ''

        launcher = self._rm.find_launcher(task)

        if not launcher:
            raise RuntimeError('no launcher found for task %s' % tid)

        self._log.debug('Launching task with %s', launcher.name)

        launch_script = '%s.launch.sh'        % tid
        exec_script   = '%s.exec.sh'          % tid
        exec_path     = '$RP_TASK_SANDBOX/%s' % exec_script

        ru.rec_makedir(sbox)

        with ru.ru_open('%s/%s' % (sbox, launch_script), 'w') as fout:

            tmp  = ''
            tmp += self._header
            tmp += self._separator
            tmp += self._get_rp_env(task)
            tmp += self._separator
            tmp += self._get_prof('launch_start', tid)

            tmp += self._separator
            tmp += '# change to task sandbox\n'
            tmp += 'cd $RP_TASK_SANDBOX\n'

            tmp += self._separator
            tmp += '# prepare launcher env\n'
            tmp += self._get_launch_env(task, launcher)

            tmp += self._separator
            tmp += '# pre-launch commands\n'
            tmp += self._get_prof('launch_pre', tid)
            tmp += self._get_pre_launch(task)

            tmp += self._separator
            tmp += '# launch commands\n'
            tmp += self._get_prof('launch_submit', tid)
            tmp += '%s\n' % self._get_launch_cmds(task, launcher, exec_path)
            tmp += 'RP_RET=$?\n'
            tmp += self._get_prof('launch_collect', tid)

            tmp += self._separator
            tmp += '# post-launch commands\n'
            tmp += self._get_prof('launch_post', tid)
            tmp += self._get_post_launch(task)

            tmp += self._separator
            tmp += self._get_prof('launch_stop', tid)
            tmp += 'exit $RP_RET\n'

            tmp += self._separator
            tmp += '\n'

            fout.write(tmp)


        ranks   = task['slots']['ranks']
        n_ranks = len(ranks)

        with ru.ru_open('%s/%s' % (sbox, exec_script), 'w') as fout:

            tmp  = ''
            tmp += self._header
            tmp += self._separator
            tmp += '# rank ID\n'
            tmp += self._get_rank_ids(n_ranks, launcher)

            tmp += self._separator
            tmp += self._get_rp_env(task)
            tmp += self._separator
            tmp += self._get_prof('exec_start', tid)

            tmp += '# task environment\n'
            tmp += self._get_task_env(task, launcher)

            tmp += self._separator
            tmp += '# pre-exec commands\n'
            tmp += self._get_prof('exec_pre', tid)
            tmp += self._get_pre_exec(task)

            # pre_rank list is applied to rank 0, dict to the ranks listed
            pre_rank = td.get('pre_rank')
            if isinstance(pre_rank, list): pre_rank = {'0': pre_rank}

            if pre_rank:
                tmp += self._separator
                tmp += self._get_prof('rank_pre', tid)
                tmp += '# pre-rank commands\n'
                tmp += 'case "$RP_RANK" in\n'
                for rank_id, cmds in pre_rank.items():
                    rank_id = int(rank_id)
                    tmp += '    %d)\n' % rank_id
                    tmp += self._get_pre_rank(cmds)
                    tmp += '        ;;\n'
                tmp += 'esac\n\n'

                tmp += self._get_rank_sync('pre_rank', n_ranks)

            tmp += self._separator
            tmp += '# execute ranks\n'
            tmp += self._get_prof('rank_start', tid)
            tmp += 'case "$RP_RANK" in\n'
            for rank_id, rank in enumerate(ranks):
                tmp += '    %d)\n' % rank_id
                tmp += self._get_rank_exec(task, rank_id, rank, launcher)
                tmp += '        ;;\n'
            tmp += 'esac\n'
            tmp += 'RP_RET=$?\n'
            tmp += self._get_prof('rank_stop', tid)

            # post_rank list is applied to rank 0, dict to the ranks listed
            post_rank = td.get('post_rank')
            if isinstance(post_rank, list): post_rank = {'0': post_rank}

            if post_rank:
                tmp += self._separator
                tmp += self._get_prof('rank_post', tid)
                tmp += self._get_rank_sync('post_rank', n_ranks)

                tmp += '\n# post-rank commands\n'
                tmp += 'case "$RP_RANK" in\n'
                for rank_id, cmds in post_rank.items():
                    rank_id = int(rank_id)
                    tmp += '    %d)\n' % rank_id
                    tmp += self._get_post_rank(cmds)
                    tmp += '        ;;\n'
                tmp += 'esac\n\n'

            tmp += self._separator
            tmp += self._get_prof('exec_post', tid)
            tmp += '# post exec commands\n'
            tmp += self._get_post_exec(task)

            tmp += self._separator
            tmp += self._get_prof('exec_stop', tid)
            tmp += 'exit $RP_RET\n'

            tmp += self._separator
            tmp += '\n'

            fout.write(tmp)


        # make sure scripts are executable
        st_l = os.stat('%s/%s' % (sbox, launch_script))
        st_e = os.stat('%s/%s' % (sbox, exec_script))
        os.chmod('%s/%s' % (sbox, launch_script), st_l.st_mode | stat.S_IEXEC)
        os.chmod('%s/%s' % (sbox, exec_script),   st_e.st_mode | stat.S_IEXEC)

        # make sure the sandbox exists
        slots_fname = '%s/%s.sl' % (sbox, tid)

        with ru.ru_open(slots_fname, 'w') as fout:
            fout.write('\n%s\n\n' % pprint.pformat(task['slots']))

        # make sure the sandbox exists
        self._prof.prof('task_mkdir', uid=tid)
        ru.rec_makedir(sbox)
        self._prof.prof('task_mkdir_done', uid=tid)

        # launch and exec script are done, get ready for execution.
        cmdline = '/bin/sh %s' % launch_script

        self._log.info('Launching task %s via %s in %s', tid, cmdline, sbox)

        _launch_out_h = ru.ru_open('%s/%s.launch.out' % (sbox, tid), 'w')

        self._prof.prof('task_run_start', uid=tid)
        task['proc'] = subprocess.Popen(args       = cmdline,
                                        executable = None,
                                        stdin      = None,
                                        stdout     = _launch_out_h,
                                        stderr     = subprocess.STDOUT,
                                        close_fds  = True,
                                        shell      = True,
                                        cwd        = sbox)
        # decoupling from parent process group is disabled,
        # in case of enabling it, one of the following options should be added:
        #    `preexec_fn=os.setsid` OR `start_new_session=True`
        self._prof.prof('task_run_ok', uid=tid)

        # store pid for last-effort termination
        _pids.append(task['proc'].pid)

        self._watch_queue.put(task)


    # --------------------------------------------------------------------------
    #
    def _watch(self):

        try:
            while not self._term.is_set():

                tasks = list()
                try:

                    # FIXME: we don't want to only wait for one Task -- then we
                    #        would pull Task state too frequently.  OTOH, we
                    #        also don't want to learn about tasks until all
                    #        slots are filled, because then we may not be able
                    #        to catch finishing tasks in time -- so there is
                    #        a fine balance here.  Balance means 100.
                    MAX_QUEUE_BULKSIZE = 100

                    while len(tasks) < MAX_QUEUE_BULKSIZE :
                        tasks.append (self._watch_queue.get_nowait())

                except queue.Empty:
                    # nothing found -- no problem, see if any tasks finished
                    pass

                # add all tasks we found to the watchlist
                for task in tasks :
                    self._tasks_to_watch.append (task)

                # check on the known tasks.
                action = self._check_running()

                if not action and not tasks :
                    # nothing happened at all!  Zzz for a bit.
                    # FIXME: make configurable
                    time.sleep(0.1)

        except Exception as e:
            self._log.exception('Error in ExecWorker watch loop (%s)' % e)
            # FIXME: this should signal the ExecWorker for shutdown...


    # --------------------------------------------------------------------------
    # Iterate over all running tasks, check their status, and decide on the
    # next step.  Also check for a requested cancellation for the tasks.
    def _check_running(self):

        action = 0
        for task in self._tasks_to_watch:

            # poll subprocess object
            exit_code = task['proc'].poll()
            tid       = task['uid']

            if exit_code is None:
                # Process is still running

                if tid in self._tasks_to_cancel:

                    # FIXME: there is a race condition between the state poll
                    # above and the kill command below.  We probably should pull
                    # state after kill again?

                    self._prof.prof('task_run_cancel_start', uid=tid)

                    # We got a request to cancel this task - send SIGTERM to the
                    # process group (which should include the actual launch
                    # method)
                  # task['proc'].kill()
                    action += 1
                    try:
                        os.killpg(task['proc'].pid, signal.SIGTERM)
                    except OSError:
                        # task is already gone, we ignore this
                        pass
                    task['proc'].wait()  # make sure proc is collected

                    with self._cancel_lock:
                        self._tasks_to_cancel.remove(tid)

                    self._prof.prof('task_run_cancel_stop', uid=tid)

                    del task['proc']  # proc is not json serializable
                    self._prof.prof('unschedule_start', uid=tid)
                    self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)
                    self.advance(task, rps.CANCELED, publish=True, push=False)

                    # we don't need to watch canceled tasks
                    self._tasks_to_watch.remove(task)

            else:

                self._prof.prof('task_run_stop', uid=tid)

                # make sure proc is collected
                task['proc'].wait()

                # we have a valid return code -- task is final
                action += 1
                self._log.info("Task %s has return code %s.", tid, exit_code)

                task['exit_code'] = exit_code

                # Free the Slots, Flee the Flots, Ree the Frots!
                self._tasks_to_watch.remove(task)
                del task['proc']  # proc is not json serializable
                self._prof.prof('unschedule_start', uid=tid)
                self.publish(rpc.AGENT_UNSCHEDULE_PUBSUB, task)

                if exit_code != 0:
                    # The task failed - fail after staging output
                    task['target_state'] = rps.FAILED

                else:
                    # The task finished cleanly, see if we need to deal with
                    # output data.  We always move to stageout, even if there
                    # are no directives -- at the very least, we'll upload
                    # stdout/stderr
                    task['target_state'] = rps.DONE

                self.advance(task, rps.AGENT_STAGING_OUTPUT_PENDING,
                                   publish=True, push=True)

        return action


    # --------------------------------------------------------------------------
    def _get_check(self, event):

        return '\\\n        || (echo "%s failed"; false) || exit 1\n' % event


    # --------------------------------------------------------------------------
    #
    # launcher
    #
# pylint: disable=unused-argument
    def _get_prof(self, event, tid, msg=''):

        return '$RP_PROF %s\n' % event


    # --------------------------------------------------------------------------
    #
    # launcher
    #
    def _get_launch_env(self, task, launcher):

        ret  = ''
        cmds = launcher.get_launcher_env()
        for cmd in cmds:
            ret += '%s %s' % (cmd, self._get_check('launcher env'))
        return ret


    # --------------------------------------------------------------------------
    #
    def _get_pre_launch(self, task):

        ret  = ''
        cmds = task['description']['pre_launch']
        for cmd in cmds:
            ret += '%s %s' % (cmd, self._get_check('pre_launch'))

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_launch_cmds(self, task, launcher, exec_path):

        ret  = '( \\\n'
        cmds = ru.as_list(launcher.get_launch_cmds(task, exec_path))
        for cmd in cmds:
            ret += '  %s \\\n' % cmd

        ret += ') 1> %s \\\n  2> %s' % (task['stdout_file_short'],
                                        task['stderr_file_short'])
        return ret


    # --------------------------------------------------------------------------
    #
    def _get_post_launch(self, task):

        ret  = ''
        cmds = task['description']['post_launch']
        for cmd in cmds:
            ret += '%s %s' % (cmd, self._get_check('post_launch'))

        return ret


    # --------------------------------------------------------------------------
    # exec
    #
    def _get_rp_env(self, task):

        tid  = task['uid']
        name = task.get('name') or tid
        sbox = os.path.realpath(task['task_sandbox_path'])

        if sbox.startswith(self._pwd):
            sbox = '$RP_PILOT_SANDBOX%s' % sbox[len(self._pwd):]

        ret  = ''
        ret += 'export RP_TASK_ID="%s"\n'          % tid
        ret += 'export RP_TASK_NAME="%s"\n'        % name
        ret += 'export RP_PILOT_ID="%s"\n'         % self._pid
        ret += 'export RP_SESSION_ID="%s"\n'       % self.sid
        ret += 'export RP_RESOURCE="%s"\n'         % self.resource
        ret += 'export RP_RESOURCE_SANDBOX="%s"\n' % self.rsbox
        ret += 'export RP_SESSION_SANDBOX="%s"\n'  % self.ssbox
        ret += 'export RP_PILOT_SANDBOX="%s"\n'    % self.psbox
        ret += 'export RP_TASK_SANDBOX="%s"\n'     % sbox
        # FIXME AM
      # ret += 'export RP_LFS="%s"\n'              % self.lfs
        ret += 'export RP_GTOD="%s"\n'             % self.gtod
        ret += 'export RP_PROF="%s"\n'             % self.prof

        if self._prof.enabled:
            ret += 'export RP_PROF_TGT="%s/%s.prof"\n' % (sbox, tid)
        else:
            ret += 'unset  RP_PROF_TGT'

        return ret


    # --------------------------------------------------------------------------
    # exec
    #
    def _get_task_env(self, task, launcher):

        ret = ''
        td  = task['description']

        # named_env's are prepared by the launcher
        if td['named_env']:
            ret += '\n# named environment\n'
            ret += '. %s\n' % launcher.get_task_named_env(td['named_env'])

        # also add any env vars requested in the task description
        if td['environment']:
            ret += '\n# task env settings\n'
            for key,val in td['environment'].items():
                ret += 'export %s="%s"\n' % (key, val)

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_pre_exec(self, task):

        ret  = ''

        cmds = task['description']['pre_exec']
        for cmd in cmds:
            ret += '%s %s' % (cmd, self._get_check('pre_exec'))

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_rank_ids(self, n_ranks, launcher):

        ret  = ''
        ret += 'export RP_RANKS=%s\n' % n_ranks
        ret += launcher.get_rank_cmd()

        if n_ranks > 1:

            # make sure that RP_RANK is known (otherwise task fails silently)
            ret += 'test -z "$RP_RANK" && echo "Cannot determine rank"\n'
            ret += 'test -z "$RP_RANK" && exit 1\n'

        else:
            ret += 'test -z "$RP_RANK" && export RP_RANK=0\n'

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_post_exec(self, task):

        ret  = ''
        cmds = task['description']['post_exec']
        for cmd in cmds:
            ret += '%s %s' % (cmd, self._get_check('post_exec'))

        return ret


    # --------------------------------------------------------------------------
    #
    # rank
    #
    def _get_pre_rank(self, cmds=None):

        ret = ''
        cmds = cmds or []
        for cmd in cmds:
            # FIXME: exit on error, but don't stall other ranks on sync
            ret += '        %s\n' % cmd

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_rank_sync(self, sig, ranks):

        # FIXME: only sync if LM needs it (`if lm._ranks_need_sync`)

        if ranks == 1:
            return ''

        # FIXME: make sure that all ranks are alive
        ret  = '# sync ranks before %s commands\n' % sig
        ret += 'echo $RP_RANK >> %s.sig\n\n' % sig
        ret += 'while test $(cat %s.sig | wc -l) -lt $RP_RANKS; do\n' % sig
        ret += '    sleep 1\n'
        ret += 'done\n\n'

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_rank_exec(self, task, rank_id, rank, launcher):

        # FIXME: this assumes that the rank has a `gpu_maps` and `core_maps`
        #        with exactly one entry, corresponding to the rank process to be
        #        started.

        ret  = ''
        gmap = rank['gpu_map']

        # FIXME: need to distinguish between logical and physical IDs
        if gmap:
            # equivalent to the 'physical' value for original `cvd_id_mode`
            gpus = ','.join([str(gpu_set[0]) for gpu_set in gmap])
            ret += '        export CUDA_VISIBLE_DEVICES=%s\n' % gpus

        cmap = rank['core_map'][0]
        ret += '        export OMP_NUM_THREADS="%d"\n' % len(cmap)

        # FIXME: core pinning goes here


        cmds = ru.as_list(launcher.get_rank_exec(task, rank_id, rank))
        for cmd in cmds:
            ret += '        %s\n' % cmd

        return ret


    # --------------------------------------------------------------------------
    #
    def _get_post_rank(self, cmds=None):

        ret = ''
        cmds = cmds or []
        for cmd in cmds:
            ret += '        %s %s' % (cmd, self._get_check('post_rank'))

        return ret


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import os

import radical.utils   as ru

from ... import states as rps

from .base import TMGRSchedulingComponent, ADDED


# the high water mark determines the percentage of task oversubscription for the
# pilots, in terms of numbers of cores
_HWM = int(os.environ.get('RADICAL_PILOT_BACKFILLING_HWM', 200))

# we consider pilots eligible for task scheduling beyond a certain start state,
# which defaults to 'PMGR_ACTIVE'.
_BF_START = os.environ.get('RADICAL_PILOT_BACKFILLING_START', rps.PMGR_ACTIVE)
_BF_STOP  = os.environ.get('RADICAL_PILOT_BACKFILLING_STOP',  rps.PMGR_ACTIVE)

_BF_START_VAL = rps._pilot_state_value(_BF_START)
_BF_STOP_VAL  = rps._pilot_state_value(_BF_STOP)


# ------------------------------------------------------------------------------
#
class Backfilling(TMGRSchedulingComponent):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        TMGRSchedulingComponent.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        self._wait_pool = dict()      # set of unscheduled tasks
        self._wait_lock = ru.RLock()  # look on the above set

        self._pids = list()
        self._idx  = 0


    # --------------------------------------------------------------------------
    #
    def add_pilots(self, pids):

      # self._log.debug('add pilots %s', pids)

        # pilots just got added.  If we did not have any pilot before, we might
        # have tasks in the wait queue waiting -- now is a good time to take
        # care of those!
        with self._wait_lock:

            # initialize custom data for the pilot
            for pid in pids:
                pilot = self._pilots[pid]['pilot']
                cores = pilot['description']['cores']
                hwm   = int(cores * _HWM / 100)
                self._pilots[pid]['info'] = {
                        'cores' : cores,
                        'hwm'   : hwm,
                        'used'  : 0,
                        'tasks' : list(),  # list of assigned task IDs
                        'done'  : list(),  # list of executed task IDs
                }

            # now we can use the pilot
            self._pids += pids
            self._schedule_tasks()


    # --------------------------------------------------------------------------
    #
    def remove_pilots(self, pids):

      # self._log.debug('rem pilots %s', pids)

        with self._pilots_lock:

            for pid in pids:

                if pid not in self._pids:
                    raise ValueError('no such pilot %s' % pid)

                self._pids.remove(pid)
                # FIXME: cancel tasks


    # --------------------------------------------------------------------------
    #
    def update_pilots(self, pids):

      # self._log.debug('update pilots for %s', pids)

        # FIXME: if PMGR_ACTIVE: schedule
        # FIXME: if FINAL:  un/re-schedule
        action = False
        with self._pilots_lock:

            for pid in pids:

                state = self._pilots[pid]['state']

              # self._log.debug('update pilot: %s %s', pid, state)

                if  rps._pilot_state_value(state) < _BF_START_VAL:
                  # self._log.debug('early')
                    # not eligible, yet
                    continue

                if  rps._pilot_state_value(state) > _BF_STOP_VAL:
                  # self._log.debug('late')
                    # not eligible anymore
                    continue

                # this pilot is eligible.  Stop checking the others, and attempt
                # reschedule
                action = True
                break

      # self._log.debug('action: %s', action)

        if action:
          # self._log.debug('upd pilot  -> schedule')
            self._schedule_tasks()


    # --------------------------------------------------------------------------
    #
    def update_tasks(self, tasks):

        self._log.debug('update tasks: %s', [u['uid'] for u in tasks])

        reschedule = False

        with self._pilots_lock, self._wait_lock:

            for task in tasks:

                uid = task['uid']

              # if uid not in self._wait_pool:
              #     continue

                state = task['state']
                pid   = task.get('pilot', '')

                self._log.debug('update  task: %s [%s] [%s]',  uid, pid, state)

                if not pid:
                    # we are not interested in state updates for unscheduled
                    # tasks
                    self._log.debug('upd task %s no pilot', uid)
                    continue

                if pid not in self._pilots:
                    # we don't handle the pilot of this task
                    self._log.debug('upd task %s not handled', uid)
                    continue

                info = self._pilots[pid]['info']

                if uid in info['done']:
                    # we don't need further state udates
                    self._log.debug('upd task %s in done', uid)
                    continue

                if  rps._task_state_value(state) <= \
                    rps._task_state_value(rps.AGENT_EXECUTING):
                    self._log.debug('upd task %s too early', uid)
                    continue

                if uid not in info['tasks']:
                    # this contradicts the task's assignment
                    self._log.debug('upd task  %s not in tasks', uid)
                    self._log.error('bf: task %s on %s inconsistent', uid, pid)
                    raise RuntimeError('inconsistent scheduler state')

                # this task is now considered done
                info['done'].append(uid)
                info['used'] -= task['description']['ranks'] \
                              * task['description']['cores_per_rank']
                reschedule = True
                self._log.debug('upd task %s - schedule (used: %s)',
                                uid, info['used'])

                if info['used'] < 0:
                    self._log.error('bf: pilot %s inconsistent', pid)
                    raise RuntimeError('inconsistent scheduler state')


        # if any pilot state was changed, consider new tasks for scheduling
        if reschedule:
            self._log.debug('upd tasks -> schedule')
            self._schedule_tasks()


    # --------------------------------------------------------------------------
    #
    def _work(self, tasks):

        with self._pilots_lock, self._wait_lock:

            for task in tasks:

                uid = task['uid']

                # not yet scheduled - put in wait pool
                self._wait_pool[uid] = task

        self._schedule_tasks()


    # --------------------------------------------------------------------------
    #
    def _schedule_tasks(self):
        """
        We have a set of tasks which we can place over a set of pilots.

        The overall objective is to keep pilots busy while load balancing across
        all pilots, even those which might yet to get added.  We achieve that
        via the following algorithm:

          - for each pilot which is being added, no matter the state:
            - assign sufficient tasks to the pilot that it can run 'n'
              generations of them, 'n' being a tunable parameter called
              'RADICAL_PILOT_BACKFILLING_HWM'.

          - for each task being completed (goes out of AGENT_EXECUTING state)
            - determine the pilot which executed it
            - backfill tasks from the wait queue until the backfilling HWM is
              reached again.

        The HWM is interpreted as percent of pilot size.  For example, a pilot
        of size 10 cores and a HWM of 200 can get tasks with a total of 20 cores
        assigned.  It can get assigned more than that, if the last task
        assigned to it surpasses the HWM.  We will not schedule any task larger
        than pilot size however.
        """

        with self._pilots_lock, self._wait_lock:

            # check if we have pilots to schedule over
            if not self._pids:
                return

            # we ignore pilots which are not yet added, are not yet in
            # BF_START_STATE, and are beyond PMGR_ACTIVE state
            pids = list()
            for pid in self._pids:

                info  = self._pilots[pid]['info']
                state = self._pilots[pid]['state']
                role  = self._pilots[pid]['role']

                if role != ADDED:
                    continue

                if  rps._pilot_state_value(state) < _BF_START_VAL:
                    # not eligible, yet
                    continue

                if  rps._pilot_state_value(state) > _BF_STOP_VAL:
                    # not ligible anymore
                    continue

                if info['used'] >= info['hwm']:
                    # pilot is full
                    continue

                pids.append(pid)

            # check if we have any eligible pilots to schedule over
            if not pids:
                return

            # cycle over available pids and add tasks until we either ran
            # out of tasks to schedule, or out of pids to schedule over

          # self._log.debug('schedule %s tasks over %s pilots',
          #         len(self._wait_pool), len(pids))

            scheduled   = list()   # tasks we want to advance
            unscheduled = dict()   # this will be the new wait pool
            for uid, task in self._wait_pool.items():

                if not pids:
                    # no more useful pilots -- move remaining tasks into
                    # unscheduled pool
                    self._log.debug(' =!= sch task  %s', uid)
                    unscheduled[uid] = task
                    continue

                cores   = task['description']['ranks'] \
                        * task['description']['cores_per_rank']
                success = False
                for pid in list(pids):

                    info = self._pilots[pid]['info']

                    if info['used'] <= info['hwm']:

                      # self._log.debug('sch task  %s -> %s', uid, pid)
                        self._log.info('schedule %s -> %s', uid, pid)

                        pilot = self._pilots[pid]['pilot']
                        info['tasks'].append(task['uid'])
                        info['used']   += cores

                        self._assign_pilot(task, pilot)
                        scheduled.append(task)
                        success = True

                        # this pilot might now be full.  If so, remove it from
                        # list of eligible pids
                        if info['used'] >= info['hwm']:
                            pids.remove(pid)

                        break  # stop looking through pilot list

                if not success:
                    # we did not find a useable pilot for this task -- keep it
                    self._log.debug(' ==! sch task  %s', uid)
                    unscheduled[uid] = task


          # self._log.debug('retain   %s tasks and  %s pilots',
          #         len(unscheduled), len(pids))

            # all unscheduled tasks *are* the new wait pool
            self._log.debug(' 1 > waits: %s', list(self._wait_pool.keys()))
            self._wait_pool = unscheduled
            self._log.debug(' 2 > waits: %s', list(self._wait_pool.keys()))

        # advance scheduled tasks
        if scheduled:
            self.advance(scheduled, rps.TMGR_STAGING_INPUT_PENDING,
                         publish=True, push=True)


        self._log.debug('\nafter schedule:')
        self._log.debug('waits:    %s', list(self._wait_pool.keys()))
      # for pid in self._pilots:
      #     print 'pilot %s' % pid
      #     pprint.pprint(self._pilots[pid]['info'])
      # self._log.debug()


# ------------------------------------------------------------------------------





__copyright__ = "Copyright 2013-2016, http://radical.rutgers.edu"
__license__   = "MIT"

import radical.utils    as ru

from ... import states as rps

from .base import TMGRSchedulingComponent


# ------------------------------------------------------------------------------
#
class RoundRobin(TMGRSchedulingComponent):

    # --------------------------------------------------------------------------
    #
    def __init__(self, cfg, session):

        TMGRSchedulingComponent.__init__(self, cfg, session)


    # --------------------------------------------------------------------------
    #
    def _configure(self):

        self._wait_pool = list()      # set of unscheduled tasks
        self._wait_lock = ru.RLock()  # look on the above set

        self._pids = list()
        self._idx  = 0

        self._log.debug('RoundRobin tmgr scheduler configured')


    # --------------------------------------------------------------------------
    #
    def add_pilots(self, pids):

        # pilots just got added.  If we did not have any pilot before, we might
        # have tasks in the wait queue waiting -- now is a good time to take
        # care of those!
        with self._wait_lock:

            self._log.debug('add pilot - waitpool %d', len(self._wait_pool))

            self._pids += pids

            if self._wait_pool:
                tasks = self._wait_pool[:]   # deep copy to avoid data recursion
                self._wait_pool = list()
                self._schedule_tasks(tasks)


    # --------------------------------------------------------------------------
    #
    def remove_pilots(self, pids):

        with self._pilots_lock:

            for pid in pids:

                if pid not in self._pids:
                    raise ValueError('no such pilot %s' % pid)

                self._pids.remove(pid)
                # FIXME: cancel tasks


    # --------------------------------------------------------------------------
    #
    def update_pilots(self, pids):

        # FIXME: we don't react on pilot state changes right now
        pass


    # --------------------------------------------------------------------------
    #
    def update_tasks(self, tasks):

        # RR scheduling is not concerned about task states
        pass


    # --------------------------------------------------------------------------
    #
    def _work(self, tasks):

        unscheduled = list()
        scheduled   = list()
        failed      = list()

        with self._pilots_lock:

            for task in tasks:

                # check if tasks are already scheduled, ie. by the application
                uid = task.get('uid')
                pid = task.get('pilot')

                self._log.debug('attempt %s', uid)

                if pid:
                    # make sure we know this pilot
                    if pid not in self._pilots:
                        self._log.error('unknown pilot %s (task %s)', uid, pid)
                        failed.append(task)
                        continue

                    pilot = self._pilots[pid]['pilot']

                    self._assign_pilot(task, pilot)
                    scheduled.append(task)

                else:
                    # not yet scheduled - put in wait pool
                    unscheduled.append(task)

        self._log.debug('failed %d / scheduled %d / unscheduled %d',
                        len(failed), len(scheduled), len(unscheduled))

        if failed     : self.advance(failed, rps.FAILED,
                                     publish=True, push=True)
        if scheduled  : self.advance(scheduled, rps.TMGR_STAGING_INPUT_PENDING,
                                     publish=True, push=True)
        if unscheduled: self._schedule_tasks(unscheduled)


    # --------------------------------------------------------------------------
    #
    def _schedule_tasks(self, tasks):

        self._log.debug('schedule %d tasks', len(tasks))

        with self._pilots_lock:

            if not self._pids:

                self._log.debug('no pilots')

                # no pilots, no schedule...
                with self._wait_lock:
                    self._wait_pool += tasks
                    return

            tasks_ok = list()
            tasks_fail = list()

            for task in tasks:

                try:
                    # determine target pilot for task
                    if self._idx >= len(self._pids):
                        self._idx = 0

                    pid   = self._pids[self._idx]
                    pilot = self._pilots[pid]['pilot']

                    self._idx += 1

                    # we assign the task to the pilot.
                    self._assign_pilot(task, pilot)

                    tasks_ok.append(task)

                except Exception:
                    self._log.exception('task schedule preparation failed')
                    tasks_fail.append(task)

            # make sure that all scheduled tasks have sandboxes known

            # advance all tasks
            self._log.debug('failed: %d, ok: %d', len(tasks_fail), len(tasks_ok))
            self.advance(tasks_fail, rps.FAILED, publish=True, push=False)
            self.advance(tasks_ok,   rps.TMGR_STAGING_INPUT_PENDING,
                         publish=True, push=True)


# ------------------------------------------------------------------------------




#!/usr/bin/env python3

import os
import time
import socket
import threading       as mt
import multiprocessing as mp

threads = list()
procs   = list()
files   = list()
sockets = list()

t_max   = 4 * 1024
p_max   = 1 * 1024
f_max   = 1 * 1024
s_max   = 1 * 1024


def _work():
    time.sleep(30)


base = '/tmp/rp_limit_%d.%%d' % os.getpid()
while True:
    try:
        f = open(base % len(files), 'w')
        files.append(f)
        if len(files) >= f_max:
            break
    except:
        break

print('files  : %5d' % len(files))
for f in files:
    os.unlink(f.name)
    f.close()

host = 'localhost'
port = 22
while True:
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host, port))
        sockets.append(s)
        if len(sockets) >= s_max:
            break
    except Exception:
        break

print('sockets: %5d' % len(sockets))
for s in sockets:
    s.close()

while True:
    try:
        t = mt.Thread(target=_work)
        t.start()
        threads.append(t)
        if len(threads) >= t_max:
            break
    except:
        break

print('threads: %5d' % len(threads))
for t in threads:
    t.join()


while True:
    try:
        p = mp.Process(target=_work)
        p.start()
        procs.append(p)
        if len(procs) >= p_max:
            break
    except:
        break

print('procs  : %5d' % len(procs))
for p in procs:
    p.join()


